##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################
import pickle
import os
from hydragnn.preprocess.utils import check_if_graph_size_variable, gather_deg
from hydragnn.utils.model import calculate_PNA_degree
from hydragnn.utils import get_comm_size_and_rank
import time
import json
from torch_geometric.utils import degree
import torch
import torch.distributed as dist


def update_config(config, train_loader, val_loader, test_loader):
    """check if config input consistent and update config with model and datasets"""

    graph_size_variable = check_if_graph_size_variable(
        train_loader, val_loader, test_loader
    )

    if "Dataset" in config:
        check_output_dim_consistent(train_loader.dataset[0], config)

    config["NeuralNetwork"] = update_config_NN_outputs(
        config["NeuralNetwork"], train_loader.dataset[0], graph_size_variable
    )

    config = normalize_output_config(config)

    config["NeuralNetwork"]["Architecture"]["input_dim"] = len(
        config["NeuralNetwork"]["Variables_of_interest"]["input_node_features"]
    )

    if config["NeuralNetwork"]["Architecture"]["model_type"] == "PNA":
        if hasattr(train_loader.dataset, "pna_deg"):
            ## Use max neighbours used in the dataset.
            deg = torch.tensor(train_loader.dataset.pna_deg)
        else:
            deg = gather_deg(train_loader.dataset)
        config["NeuralNetwork"]["Architecture"]["pna_deg"] = deg.tolist()
        config["NeuralNetwork"]["Architecture"]["max_neighbours"] = len(deg) - 1
    else:
        config["NeuralNetwork"]["Architecture"]["pna_deg"] = None

    if "radius" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["radius"] = None
    if "num_gaussians" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["num_gaussians"] = None
    if "num_filters" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["num_filters"] = None
    if "envelope_exponent" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["envelope_exponent"] = None
    if "num_after_skip" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["num_after_skip"] = None
    if "num_before_skip" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["num_before_skip"] = None
    if "basis_emb_size" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["basis_emb_size"] = None
    if "int_emb_size" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["int_emb_size"] = None
    if "out_emb_size" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["out_emb_size"] = None
    if "num_radial" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["num_radial"] = None
    if "num_spherical" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["num_spherical"] = None

    config["NeuralNetwork"]["Architecture"] = update_config_edge_dim(
        config["NeuralNetwork"]["Architecture"]
    )

    config["NeuralNetwork"]["Architecture"] = update_config_equivariance(
        config["NeuralNetwork"]["Architecture"]
    )

    if "freeze_conv_layers" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["freeze_conv_layers"] = False
    if "initial_bias" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["initial_bias"] = None

    if "Optimizer" not in config["NeuralNetwork"]["Training"]:
        config["NeuralNetwork"]["Training"]["Optimizer"]["type"] = "AdamW"

    if "loss_function_type" not in config["NeuralNetwork"]["Training"]:
        config["NeuralNetwork"]["Training"]["loss_function_type"] = "mse"

    if "activation_function" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["activation_function"] = "relu"

    if "SyncBatchNorm" not in config["NeuralNetwork"]["Architecture"]:
        config["NeuralNetwork"]["Architecture"]["SyncBatchNorm"] = False
    return config


def update_config_equivariance(config):
    equivariant_models = ["EGNN", "SchNet"]
    if "equivariance" in config and config["equivariance"]:
        assert (
            config["model_type"] in equivariant_models
        ), "E(3) equivariance can only be ensured for EGNN and SchNet."
    elif "equivariance" not in config:
        config["equivariance"] = False
    return config


def update_config_edge_dim(config):
    config["edge_dim"] = None
    edge_models = ["PNA", "CGCNN", "SchNet", "EGNN"]
    if "edge_features" in config and config["edge_features"]:
        assert (
            config["model_type"] in edge_models
        ), "Edge features can only be used with EGNN, SchNet, PNA and CGCNN."
        config["edge_dim"] = len(config["edge_features"])
    elif config["model_type"] == "CGCNN":
        # CG always needs an integer edge_dim
        # PNA would fail with integer edge_dim without edge_attr
        config["edge_dim"] = 0
    return config


def check_output_dim_consistent(data, config):
    output_type = config["NeuralNetwork"]["Variables_of_interest"]["type"]
    output_index = config["NeuralNetwork"]["Variables_of_interest"]["output_index"]
    if hasattr(data, "y_loc"):
        for ihead in range(len(output_type)):
            if output_type[ihead] == "graph":
                assert (
                    data.y_loc[0, ihead + 1].item() - data.y_loc[0, ihead].item()
                    == config["Dataset"]["graph_features"]["dim"][output_index[ihead]]
                )
            elif output_type[ihead] == "node":
                assert (
                    data.y_loc[0, ihead + 1].item() - data.y_loc[0, ihead].item()
                ) // data.num_nodes == config["Dataset"]["node_features"]["dim"][
                    output_index[ihead]
                ]


def update_config_NN_outputs(config, data, graph_size_variable):
    """ "Extract architecture output dimensions and set node-level prediction architecture"""

    output_type = config["Variables_of_interest"]["type"]
    if hasattr(data, "y_loc"):
        dims_list = []
        for ihead in range(len(output_type)):
            if output_type[ihead] == "graph":
                dim_item = data.y_loc[0, ihead + 1].item() - data.y_loc[0, ihead].item()
            elif output_type[ihead] == "node":
                if (
                    graph_size_variable
                    and config["Architecture"]["output_heads"]["node"]["type"]
                    == "mlp_per_node"
                ):
                    raise ValueError(
                        '"mlp_per_node" is not allowed for variable graph size, Please set config["NeuralNetwork"]["Architecture"]["output_heads"]["node"]["type"] to be "mlp" or "conv" in input file.'
                    )
                dim_item = (
                    data.y_loc[0, ihead + 1].item() - data.y_loc[0, ihead].item()
                ) // data.num_nodes
            else:
                raise ValueError("Unknown output type", output_type[ihead])
            dims_list.append(dim_item)
    else:
        for ihead in range(len(output_type)):
            if output_type[ihead] != "graph":
                raise ValueError(
                    "y_loc is needed for outputs that are not at graph levels",
                    output_type[ihead],
                )
        dims_list = config["Variables_of_interest"]["output_dim"]

    config["Architecture"]["output_dim"] = dims_list
    config["Architecture"]["output_type"] = output_type
    config["Architecture"]["num_nodes"] = data.num_nodes
    return config


def normalize_output_config(config):
    var_config = config["NeuralNetwork"]["Variables_of_interest"]
    if "denormalize_output" in var_config and var_config["denormalize_output"]:
        if (
            var_config.get("minmax_node_feature") is not None
            and var_config.get("minmax_graph_feature") is not None
        ):
            dataset_path = None
        ###loading min/max values from input data file. Only one path is needed
        elif list(config["Dataset"]["path"].values())[0].endswith(".pkl"):
            dataset_path = list(config["Dataset"]["path"].values())[0]
        else:
            if "total" in config["Dataset"]["path"].keys():
                dataset_path = f"{os.environ['SERIALIZED_DATA_PATH']}/serialized_dataset/{config['Dataset']['name']}.pkl"
            else:
                dataset_path = f"{os.environ['SERIALIZED_DATA_PATH']}/serialized_dataset/{config['Dataset']['name']}_train.pkl"
        var_config = update_config_minmax(dataset_path, var_config)
    else:
        var_config["denormalize_output"] = False

    config["NeuralNetwork"]["Variables_of_interest"] = var_config
    return config


def update_config_minmax(dataset_path, config):
    """load minimum and maximum values from dataset_path, if need denormalize,"""
    ## Check first if "minmax_graph_feature" and "minmax_graph_feature"
    if "minmax_node_feature" not in config and "minmax_graph_feature" not in config:
        with open(dataset_path, "rb") as f:
            node_minmax = pickle.load(f)
            graph_minmax = pickle.load(f)
    else:
        node_minmax = config["minmax_node_feature"]
        graph_minmax = config["minmax_graph_feature"]
    config["x_minmax"] = []
    config["y_minmax"] = []
    feature_indices = [i for i in config["input_node_features"]]
    for item in feature_indices:
        config["x_minmax"].append(node_minmax[:, item].tolist())
    output_type = config["type"]
    output_index = config["output_index"]
    for item in range(len(output_type)):
        if output_type[item] == "graph":
            config["y_minmax"].append(graph_minmax[:, output_index[item]].tolist())
        elif output_type[item] == "node":
            config["y_minmax"].append(node_minmax[:, output_index[item]].tolist())
        else:
            raise ValueError("Unknown output type", output_type[item])
    return config


def get_log_name_config(config):
    return (
        config["NeuralNetwork"]["Architecture"]["model_type"]
        + "-r-"
        + str(config["NeuralNetwork"]["Architecture"]["radius"])
        + "-ncl-"
        + str(config["NeuralNetwork"]["Architecture"]["num_conv_layers"])
        + "-hd-"
        + str(config["NeuralNetwork"]["Architecture"]["hidden_dim"])
        + "-ne-"
        + str(config["NeuralNetwork"]["Training"]["num_epoch"])
        + "-lr-"
        + str(config["NeuralNetwork"]["Training"]["Optimizer"]["learning_rate"])
        + "-bs-"
        + str(config["NeuralNetwork"]["Training"]["batch_size"])
        + "-data-"
        + config["Dataset"]["name"][
            : config["Dataset"]["name"].rfind("_")
            if config["Dataset"]["name"].rfind("_") > 0
            else None
        ]
        + "-node_ft-"
        + "".join(
            str(x)
            for x in config["NeuralNetwork"]["Variables_of_interest"][
                "input_node_features"
            ]
        )
        + "-task_weights-"
        + "".join(
            str(weigh) + "-"
            for weigh in config["NeuralNetwork"]["Architecture"]["task_weights"]
        )
    )


def save_config(config, log_name, path="./logs/"):
    """Save config"""
    _, world_rank = get_comm_size_and_rank()
    if world_rank == 0:
        fname = os.path.join(path, log_name, "config.json")
        with open(fname, "w") as f:
            json.dump(config, f)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import numpy as np

import torch
from torch_geometric.data import Data
from torch import tensor

from ase.io.cfg import read_cfg

from hydragnn.preprocess.raw_dataset_loader import AbstractRawDataLoader

# WARNING: DO NOT use collective communication calls here because only rank 0 uses this routines


class CFG_RawDataLoader(AbstractRawDataLoader):
    """A class used for loading raw files that contain data representing atom structures, transforms it and stores the structures as file of serialized structures.
    Most of the class methods are hidden, because from outside a caller needs only to know about
    load_raw_data method.

    Methods
    -------
    load_raw_data()
        Loads the raw files from specified path, performs the transformation to Data objects and normalization of values.
    """

    def __init__(self, config, dist=False):
        super(CFG_RawDataLoader, self).__init__(config, dist)

    def transform_input_to_data_object_base(self, filepath):
        data_object = self.__transform_CFG_input_to_data_object_base(filepath=filepath)
        return data_object

    def __transform_CFG_input_to_data_object_base(self, filepath):
        """Transforms lines of strings read from the raw data CFG file to Data object and returns it.

        Parameters
        ----------
        lines:
          content of data file with all the graph information
        Returns
        ----------
        Data
            Data object representing structure of a graph sample.
        """

        if filepath.endswith(".cfg"):

            data_object = self.__transform_ASE_object_to_data_object(filepath)

            return data_object

        else:
            return None

    def __transform_ASE_object_to_data_object(self, filepath):

        # FIXME:
        #  this still assumes bulk modulus is specific to the CFG format.
        #  To deal with multiple files across formats, one should generalize this function
        #  by moving the reading of the .bulk file in a standalone routine.
        #  Morevoer, this approach assumes tha there is only one global feature to look at,
        #  and that this global feature is specicially retrieveable in a file with the string *bulk* inside.

        ase_object = read_cfg(filepath)

        data_object = Data()

        data_object.supercell_size = tensor(ase_object.cell.array).float()
        data_object.pos = tensor(ase_object.arrays["positions"]).float()
        proton_numbers = np.expand_dims(ase_object.arrays["numbers"], axis=1)
        masses = np.expand_dims(ase_object.arrays["masses"], axis=1)
        c_peratom = np.expand_dims(ase_object.arrays["c_peratom"], axis=1)
        fx = np.expand_dims(ase_object.arrays["fx"], axis=1)
        fy = np.expand_dims(ase_object.arrays["fy"], axis=1)
        fz = np.expand_dims(ase_object.arrays["fz"], axis=1)
        node_feature_matrix = np.concatenate(
            (proton_numbers, masses, c_peratom, fx, fy, fz), axis=1
        )
        data_object.x = tensor(node_feature_matrix).float()

        filename_without_extension = os.path.splitext(filepath)[0]

        if os.path.exists(os.path.join(filename_without_extension + ".bulk")):
            filename_bulk = os.path.join(filename_without_extension + ".bulk")
            f = open(filename_bulk, "r", encoding="utf-8")
            lines = f.readlines()
            graph_feat = lines[0].split(None, 2)
            g_feature = []
            # collect graph features
            for item in range(len(self.graph_feature_dim)):
                for icomp in range(self.graph_feature_dim[item]):
                    it_comp = self.graph_feature_col[item] + icomp
                    g_feature.append(float(graph_feat[it_comp].strip()))
            data_object.y = tensor(g_feature)

        return data_object



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################
import os, json
import torch
import random
import hydragnn
from .test_graphs import unittest_train_model


def unittest_model_prediction(config):
    _, _ = hydragnn.utils.distributed.setup_ddp()

    (
        train_loader,
        val_loader,
        test_loader,
    ) = hydragnn.preprocess.load_data.dataset_loading_and_splitting(config=config)

    model = hydragnn.models.create.create_model_config(
        config=config["NeuralNetwork"],
        verbosity=config["Verbosity"]["level"],
    )

    model = hydragnn.utils.get_distributed_model(model, config["Verbosity"]["level"])

    log_name = hydragnn.utils.config_utils.get_log_name_config(config)
    hydragnn.utils.model.load_existing_model(model, log_name)

    model.eval()
    # two checkings
    # 1. entire test set
    thresholds = [0.2]
    _, _, true_values, predicted_values = hydragnn.train.test(
        test_loader, model, config["Verbosity"]["level"]
    )
    # 2.check a random selected sample
    isample = random.randrange(len(test_loader.dataset))
    graph_sample = test_loader.dataset[isample].to(model.device)
    true_sample = graph_sample.y.squeeze()
    predicted_sample = model(graph_sample)
    yloc = graph_sample.y_loc.squeeze()

    mae = torch.nn.L1Loss()
    for ihead in range(model.module.num_heads):
        head_true = true_values[ihead]
        head_pred = predicted_values[ihead]
        test_mae = mae(head_true, head_pred)
        print("For head ", ihead, "; MAE of test set =", test_mae)
        assert test_mae < thresholds[0], "MAE sample checking failed for test set!"


# test loading and predictiong of a saved model from previous training
def pytest_model_loadpred():
    model_type = "PNA"
    ci_input = "ci_multihead.json"
    config_file = os.path.join(os.getcwd(), "tests/inputs", ci_input)
    with open(config_file, "r") as f:
        config = json.load(f)
    config["NeuralNetwork"]["Architecture"]["model_type"] = model_type
    # get the directory of trained model
    log_name = hydragnn.utils.config_utils.get_log_name_config(config)
    modelfile = os.path.join("./logs/", log_name, log_name + ".pk")
    # check if pretrained model and pkl dataset files exists
    case_exist = True
    config_file = os.path.join("./logs/", log_name, "config.json")
    if not (os.path.isfile(modelfile) and os.path.isfile(config_file)):
        print("Model or configure file not found: ", modelfile, config_file)
        case_exist = False
    else:
        with open(config_file, "r") as f:
            config = json.load(f)
        for dataset_name, raw_data_path in config["Dataset"]["path"].items():
            if not os.path.isfile(raw_data_path):
                print(dataset_name, "dataset not found: ", raw_data_path)
                case_exist = False
                break
    if not case_exist:
        unittest_train_model(
            config["NeuralNetwork"]["Architecture"]["model_type"],
            "ci_multihead.json",
            False,
            False,
        )
    unittest_model_prediction(config)



from .Base import Base
from .GATStack import GATStack
from .GINStack import GINStack
from .PNAStack import PNAStack
from .GINStack import GINStack
from .create import create_model, create_model_config



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os, json
import pytest

import shutil

import hydragnn, tests


# Loss function unit test called by pytest wrappers.
# Note the intent of this test is to make sure all interfaces work - it does not assert anything
def unittest_loss_and_activation_functions(
    activation_function_type, loss_function_type, ci_input, overwrite_data=False
):
    world_size, rank = hydragnn.utils.get_comm_size_and_rank()

    os.environ["SERIALIZED_DATA_PATH"] = os.getcwd()

    # Read in config settings and override model type.
    config_file = os.path.join(os.getcwd(), "tests/inputs", ci_input)
    with open(config_file, "r") as f:
        config = json.load(f)

    config["NeuralNetwork"]["Architecture"][
        "activation_function"
    ] = activation_function_type

    # use pkl files if exist by default
    for dataset_name in config["Dataset"]["path"].keys():
        if dataset_name == "total":
            pkl_file = (
                os.environ["SERIALIZED_DATA_PATH"]
                + "/serialized_dataset/"
                + config["Dataset"]["name"]
                + ".pkl"
            )
        else:
            pkl_file = (
                os.environ["SERIALIZED_DATA_PATH"]
                + "/serialized_dataset/"
                + config["Dataset"]["name"]
                + "_"
                + dataset_name
                + ".pkl"
            )
        if os.path.exists(pkl_file):
            config["Dataset"]["path"][dataset_name] = pkl_file

    if rank == 0:
        num_samples_tot = 500
        # check if serialized pickle files or folders for raw files provided
        pkl_input = False
        if list(config["Dataset"]["path"].values())[0].endswith(".pkl"):
            pkl_input = True
        # only generate new datasets, if not pkl
        if not pkl_input:
            for dataset_name, data_path in config["Dataset"]["path"].items():
                if overwrite_data:
                    shutil.rmtree(data_path)
                if not os.path.exists(data_path):
                    os.makedirs(data_path)
                if dataset_name == "total":
                    num_samples = num_samples_tot
                elif dataset_name == "train":
                    num_samples = int(
                        num_samples_tot
                        * config["NeuralNetwork"]["Training"]["perc_train"]
                    )
                elif dataset_name == "test":
                    num_samples = int(
                        num_samples_tot
                        * (1 - config["NeuralNetwork"]["Training"]["perc_train"])
                        * 0.5
                    )
                elif dataset_name == "validate":
                    num_samples = int(
                        num_samples_tot
                        * (1 - config["NeuralNetwork"]["Training"]["perc_train"])
                        * 0.5
                    )
                if not os.listdir(data_path):
                    tests.deterministic_graph_data(
                        data_path, number_configurations=num_samples
                    )

    config["NeuralNetwork"]["Training"]["loss_function_type"] = loss_function_type
    config["NeuralNetwork"]["Training"]["num_epoch"] = 2

    # Make sure training works with each loss function.
    hydragnn.run_training(config)


# Test all supported loss function types. Separate input file because only 2 steps are run.
@pytest.mark.parametrize("loss_function_type", ["mse", "mae", "rmse"])
def pytest_loss_functions(loss_function_type, ci_input="ci.json", overwrite_data=False):
    unittest_loss_and_activation_functions(
        "relu", loss_function_type, ci_input, overwrite_data
    )


# Test all supported activation function types.
@pytest.mark.parametrize(
    "activation_function_type",
    ["relu", "selu", "prelu", "elu", "lrelu_01", "lrelu_025", "lrelu_05"],
)
def pytest_activation_functions_multihead(
    activation_function_type, ci_input="ci_multihead.json", overwrite_data=False
):
    unittest_loss_and_activation_functions(
        activation_function_type, "mse", ci_input, overwrite_data
    )


# Test all supported activation function types.
@pytest.mark.parametrize(
    "activation_function_type",
    ["relu", "selu", "prelu", "elu", "lrelu_01", "lrelu_025", "lrelu_05"],
)
def pytest_activation_functions_vectoroutput(
    activation_function_type, ci_input="ci_vectoroutput.json", overwrite_data=False
):
    unittest_loss_and_activation_functions(
        activation_function_type, "mse", ci_input, overwrite_data
    )



import os
import shutil
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt


def find_bin(comp, nbins):
    bins = np.linspace(0, 1, nbins)
    for bi in range(len(bins) - 1):
        if comp > bins[bi] and comp < bins[bi + 1]:
            return bi
    return nbins - 1


def compositional_histogram_cutoff(
    dir,
    elements_list,
    histogram_cutoff,
    num_bins,
    overwrite_data=False,
    create_plots=True,
):
    """
    Downselect LSMS data with maximum number of samples per binary composition.
    """

    if dir.endswith("/"):
        dir = dir[:-1]
    new_dir = dir + "_histogram_cutoff/"

    if os.path.exists(new_dir):
        if overwrite_data:
            shutil.rmtree(new_dir)
        else:
            print("Exiting: path to histogram cutoff data already exists")
            return
    if not os.path.exists(new_dir):
        os.makedirs(new_dir)

    comp_final = []
    comp_all = np.zeros([num_bins])
    for filename in tqdm(os.listdir(dir)):

        path = os.path.join(dir, filename)
        # This is LSMS specific - it assumes only one header line and only atoms following.
        atoms = np.loadtxt(path, skiprows=1)

        elements, counts = np.unique(atoms[:, 0], return_counts=True)

        # Fixup for the pure component cases.
        for e, elem in enumerate(elements_list):
            if elem not in elements:
                elements = np.insert(elements, e, elem)
                counts = np.insert(counts, e, 0)

        num_atoms = atoms.shape[0]
        composition = counts[0] / num_atoms

        b = find_bin(composition, num_bins)
        comp_all[b] += 1
        if comp_all[b] < histogram_cutoff:
            comp_final.append(composition)
            new_path = os.path.join(new_dir, filename)
            os.symlink(path, new_path)

    if create_plots:
        plt.figure(0)
        plt.hist(comp_final, bins=num_bins)
        plt.savefig("composition_histogram_cutoff.png")

        plt.figure(1)
        w = 1 / num_bins
        plt.bar(np.linspace(0, 1, num_bins), comp_all, width=w)
        plt.savefig("composition_initial.png")



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import torch
import torch.nn.functional as F
from torch.nn import ModuleList
from torch_geometric.nn import CGConv, BatchNorm, global_mean_pool, Sequential
from .Base import Base


class CGCNNStack(Base):
    def __init__(
        self,
        edge_dim: int,
        input_dim,
        output_dim,
        *args,
        **kwargs,
    ):
        self.edge_dim = edge_dim

        # CGCNN does not change embedding dimensions
        # We use input dimension (first argument of base constructor)
        #    also as hidden dimension (second argument of base constructor)
        # We therefore pass all required args explicitly.
        super().__init__(
            input_dim,
            input_dim,
            output_dim,
            *args,
            **kwargs,
        )

    def get_conv(self, input_dim, _):
        cgcnn = CGConv(
            channels=input_dim,
            dim=self.edge_dim,
            aggr="add",
            batch_norm=False,
            bias=True,
        )

        input_args = "x, pos, edge_index"
        conv_args = "x, edge_index"

        if self.use_edge_attr:
            input_args += ", edge_attr"
            conv_args += ", edge_attr"

        return Sequential(
            input_args,
            [
                (cgcnn, conv_args + " -> x"),
                (lambda x, pos: [x, pos], "x, pos -> x, pos"),
            ],
        )

    def _init_node_conv(self):
        """It overwrites _init_node_conv() in Base since purely convolutional layers in _init_node_conv() is not implemented yet.
        Here it serves as a temporary place holder. Purely cgcnn conv is not feasible for node feature predictions with
        arbitrary output dimensions, unless we combine it with mlp"""
        # *******convolutional layers for node level predictions******* #
        node_feature_ind = [
            i for i, head_type in enumerate(self.head_type) if head_type == "node"
        ]
        if len(node_feature_ind) == 0:
            return
        self.num_conv_layers_node = self.config_heads["node"]["num_headlayers"]
        self.hidden_dim_node = self.config_heads["node"]["dim_headlayers"]
        # fixme: CGConv layer alone will present the same out dimension with the input, instead of having different "in_channels" and "out_channels" as in the other conv layers;
        # so to predict output node features with different dimensions from the input node feature's, CGConv can be
        # combined with, e.g.,mlp
        for ihead in range(self.num_heads):
            if (
                self.head_type[ihead] == "node"
                and self.config_heads["node"]["type"] == "conv"
            ):
                raise ValueError(
                    '"conv" for node features decoder part in CGCNN is not ready yet. Please set config["NeuralNetwork"]["Architecture"]["output_heads"]["node"]["type"] to be "mlp" or "mlp_per_node" in input file.'
                )

    def __str__(self):
        return "CGCNNStack"



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import torch
from torch_geometric.data import Data

from hydragnn.models.GINStack import GINStack
from hydragnn.models.PNAStack import PNAStack
from hydragnn.models.GATStack import GATStack
from hydragnn.models.MFCStack import MFCStack
from hydragnn.models.CGCNNStack import CGCNNStack
from hydragnn.models.SAGEStack import SAGEStack
from hydragnn.models.SCFStack import SCFStack
from hydragnn.models.DIMEStack import DIMEStack
from hydragnn.models.EGCLStack import EGCLStack

from hydragnn.utils.distributed import get_device
from hydragnn.utils.print_utils import print_distributed
from hydragnn.utils.time_utils import Timer


def create_model_config(
    config: dict,
    verbosity: int = 0,
    use_gpu: bool = True,
):
    return create_model(
        config["Architecture"]["model_type"],
        config["Architecture"]["input_dim"],
        config["Architecture"]["hidden_dim"],
        config["Architecture"]["output_dim"],
        config["Architecture"]["output_type"],
        config["Architecture"]["output_heads"],
        config["Architecture"]["activation_function"],
        config["Training"]["loss_function_type"],
        config["Architecture"]["task_weights"],
        config["Architecture"]["num_conv_layers"],
        config["Architecture"]["freeze_conv_layers"],
        config["Architecture"]["initial_bias"],
        config["Architecture"]["num_nodes"],
        config["Architecture"]["max_neighbours"],
        config["Architecture"]["edge_dim"],
        config["Architecture"]["pna_deg"],
        config["Architecture"]["num_before_skip"],
        config["Architecture"]["num_after_skip"],
        config["Architecture"]["num_radial"],
        config["Architecture"]["basis_emb_size"],
        config["Architecture"]["int_emb_size"],
        config["Architecture"]["out_emb_size"],
        config["Architecture"]["envelope_exponent"],
        config["Architecture"]["num_spherical"],
        config["Architecture"]["num_gaussians"],
        config["Architecture"]["num_filters"],
        config["Architecture"]["radius"],
        config["Architecture"]["equivariance"],
        verbosity,
        use_gpu,
    )


# FIXME: interface does not include ilossweights_hyperp, ilossweights_nll, dropout
def create_model(
    model_type: str,
    input_dim: int,
    hidden_dim: int,
    output_dim: list,
    output_type: list,
    output_heads: dict,
    activation_function: str,
    loss_function_type: str,
    task_weights: list,
    num_conv_layers: int,
    freeze_conv: bool = False,
    initial_bias: float = None,
    num_nodes: int = None,
    max_neighbours: int = None,
    edge_dim: int = None,
    pna_deg: torch.tensor = None,
    num_before_skip: int = None,
    num_after_skip: int = None,
    num_radial: int = None,
    basis_emb_size: int = None,
    int_emb_size: int = None,
    out_emb_size: int = None,
    envelope_exponent: int = None,
    num_spherical: int = None,
    num_gaussians: int = None,
    num_filters: int = None,
    radius: float = None,
    equivariance: bool = False,
    verbosity: int = 0,
    use_gpu: bool = True,
):
    timer = Timer("create_model")
    timer.start()
    torch.manual_seed(0)

    device = get_device(use_gpu, verbosity_level=verbosity)

    # Note: model-specific inputs must come first.
    if model_type == "GIN":
        model = GINStack(
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "PNA":
        assert pna_deg is not None, "PNA requires degree input."
        model = PNAStack(
            pna_deg,
            edge_dim,
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "GAT":
        # FIXME: expose options to users
        heads = 6
        negative_slope = 0.05
        model = GATStack(
            heads,
            negative_slope,
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "MFC":
        assert max_neighbours is not None, "MFC requires max_neighbours input."
        model = MFCStack(
            max_neighbours,
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "CGCNN":
        model = CGCNNStack(
            edge_dim,
            input_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "SAGE":
        model = SAGEStack(
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "SchNet":
        assert num_gaussians is not None, "SchNet requires num_guassians input."
        assert num_filters is not None, "SchNet requires num_filters input."
        assert radius is not None, "SchNet requires radius input."
        model = SCFStack(
            num_gaussians,
            num_filters,
            radius,
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            max_neighbours=max_neighbours,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "DimeNet":
        assert basis_emb_size is not None, "DimeNet requires basis_emb_size input."
        assert (
            envelope_exponent is not None
        ), "DimeNet requires envelope_exponent input."
        assert int_emb_size is not None, "DimeNet requires int_emb_size input."
        assert out_emb_size is not None, "DimeNet requires out_emb_size input."
        assert num_after_skip is not None, "DimeNet requires num_after_skip input."
        assert num_before_skip is not None, "DimeNet requires num_before_skip input."
        assert num_radial is not None, "DimeNet requires num_radial input."
        assert num_spherical is not None, "DimeNet requires num_spherical input."
        assert radius is not None, "DimeNet requires radius input."
        model = DIMEStack(
            basis_emb_size,
            envelope_exponent,
            int_emb_size,
            out_emb_size,
            num_after_skip,
            num_before_skip,
            num_radial,
            num_spherical,
            radius,
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            max_neighbours=max_neighbours,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    elif model_type == "EGNN":
        model = EGCLStack(
            edge_dim,
            input_dim,
            hidden_dim,
            output_dim,
            output_type,
            output_heads,
            activation_function,
            loss_function_type,
            equivariance,
            max_neighbours=max_neighbours,
            loss_weights=task_weights,
            freeze_conv=freeze_conv,
            initial_bias=initial_bias,
            num_conv_layers=num_conv_layers,
            num_nodes=num_nodes,
        )

    else:
        raise ValueError("Unknown model_type: {0}".format(model_type))

    timer.stop()

    return model.to(device)



""" 
This is a dummy package for gptl4py (https://github.com/jychoi-hpc/gptl4py).
When gptl4py is not available, use as follows:
```
try:
    import gptl4py as gp
except ImportError:
    import gptl4py_dummy as gp
```
"""

from __future__ import absolute_import
from functools import wraps
from contextlib import contextmanager
import torch.cuda.nvtx as nvtx


def initialize():
    pass


def finalize():
    pass


def start(name):
    pass


def stop(name):
    pass


def pr_file(filename):
    pass


def pr_summary_file(filename):
    pass


def profile(x_or_func=None, *decorator_args, **decorator_kws):
    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kws):
            if "x_or_func" not in locals() or callable(x_or_func) or x_or_func is None:
                x = func.__name__
            else:
                x = x_or_func
            start(x)
            out = func(*args, **kws)
            stop(x)
            return out

        return wrapper

    return _decorator(x_or_func) if callable(x_or_func) else _decorator


@contextmanager
def nvtx_timer(x):
    start(x)
    yield
    stop(x)



from .print_utils import print_distributed, iterate_tqdm, setup_log
from .distributed import (
    get_comm_size_and_rank,
    get_device_list,
    get_device,
    get_device_name,
    get_device_from_name,
    is_model_distributed,
    get_distributed_model,
    setup_ddp,
    nsplit,
    comm_reduce,
)
from .model import (
    save_model,
    get_summary_writer,
    unsorted_segment_mean,
    load_existing_model,
    load_existing_model_config,
    loss_function_selection,
    tensor_divide,
    EarlyStopping,
)
from .time_utils import Timer, print_timers
from .config_utils import (
    update_config,
    update_config_minmax,
    get_log_name_config,
    save_config,
)

from .optimizer import select_optimizer
from .atomicdescriptors import atomicdescriptors



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import pytest

import subprocess


@pytest.mark.parametrize("example", ["qm9", "md17"])
@pytest.mark.mpi_skip()
def pytest_examples(example):
    path = os.path.join(os.path.dirname(__file__), "..", "examples", example)
    file_path = os.path.join(path, example + ".py")
    return_code = subprocess.call(["python", file_path])

    # Check the file ran without error.
    assert return_code == 0



from torch import tensor
from torch_geometric.data import Data
from hydragnn.utils.abstractrawdataset import AbstractRawDataset


class LSMSDataset(AbstractRawDataset):
    def __init__(self, config, dist=False, sampling=None):
        super().__init__(config, dist, sampling)

    def transform_input_to_data_object_base(self, filepath):
        data_object = self.__transform_LSMS_input_to_data_object_base(filepath=filepath)

        return data_object

    def __transform_LSMS_input_to_data_object_base(self, filepath):
        """Transforms lines of strings read from the raw data LSMS file to Data object and returns it.

        Parameters
        ----------
        lines:
          content of data file with all the graph information
        Returns
        ----------
        Data
            Data object representing structure of a graph sample.
        """

        data_object = Data()

        f = open(filepath, "r", encoding="utf-8")

        lines = f.readlines()
        graph_feat = lines[0].split(None, 2)
        g_feature = []
        # collect graph features
        for item in range(len(self.graph_feature_dim)):
            for icomp in range(self.graph_feature_dim[item]):
                it_comp = self.graph_feature_col[item] + icomp
                g_feature.append(float(graph_feat[it_comp].strip()))
        data_object.y = tensor(g_feature)

        node_feature_matrix = []
        node_position_matrix = []
        for line in lines[1:]:
            node_feat = line.split(None, 11)

            x_pos = float(node_feat[2].strip())
            y_pos = float(node_feat[3].strip())
            z_pos = float(node_feat[4].strip())
            node_position_matrix.append([x_pos, y_pos, z_pos])

            node_feature = []
            for item in range(len(self.node_feature_dim)):
                for icomp in range(self.node_feature_dim[item]):
                    it_comp = self.node_feature_col[item] + icomp
                    node_feature.append(float(node_feat[it_comp].strip()))
            node_feature_matrix.append(node_feature)

        f.close()

        data_object.pos = tensor(node_position_matrix)
        data_object.x = tensor(node_feature_matrix)
        data_object = self.__charge_density_update_for_LSMS(data_object)
        return data_object

    def __charge_density_update_for_LSMS(self, data_object: Data):
        """Calculate charge density for LSMS format
        Parameters
        ----------
        data_object: Data
            Data object representing structure of a graph sample.

        Returns
        ----------
        Data
            Data object representing structure of a graph sample.
        """
        num_of_protons = data_object.x[:, 0]
        charge_density = data_object.x[:, 1]
        charge_density -= num_of_protons
        data_object.x[:, 1] = charge_density
        return data_object



import os
import pickle

from .print_utils import log

from hydragnn.utils import get_comm_size_and_rank
from hydragnn.utils.abstractbasedataset import AbstractBaseDataset


class SerializedDataset(AbstractBaseDataset):
    """Serialized Dataset"""

    def __init__(self, basedir, datasetname, label, dist=False):
        """
        Parameters
        ----------
        basedir: basedir
        datasetname: dataset name
        label: label
        """
        super().__init__()

        self.basedir = basedir
        self.datasetname = datasetname
        self.label = label
        self.dist = dist

        if dist:
            _, rank = get_comm_size_and_rank()
            basename = "%s-%s-%d.pkl" % (datasetname, label, rank)
        else:
            basename = "%s-%s.pkl" % (datasetname, label)

        fname = os.path.join(basedir, basename)
        with open(fname, "rb") as f:
            self.minmax_node_feature = pickle.load(f)
            self.minmax_graph_feature = pickle.load(f)
            self.dataset = pickle.load(f)

        log("Pickle files:", self.label, len(self.dataset))

    def len(self):
        return len(self.dataset)

    def get(self, i):
        return self.dataset[i]


class SerializedWriter:
    """Serialized Dataset Writer"""

    def __init__(
        self,
        dataset,
        basedir,
        datasetname,
        label="total",
        minmax_node_feature=None,
        minmax_graph_feature=None,
        dist=False,
    ):
        """
        Parameters
        ----------
        dataset: locally owned dataset (should be iterable)
        basedir: basedir
        datasetname: dataset name
        label: label
        nmax: nmax in case of subdir
        minmax_node_feature: minmax_node_feature
        minmax_graph_feature: minmax_graph_feature
        dist: distributed or not
        """
        if not os.path.exists(basedir):
            os.makedirs(basedir)

        if dist:
            _, rank = get_comm_size_and_rank()
            basename = "%s-%s-%d.pkl" % (datasetname, label, rank)
        else:
            basename = "%s-%s.pkl" % (datasetname, label)

        fname = os.path.join(basedir, basename)
        with open(fname, "wb") as f:
            pickle.dump(minmax_node_feature, f)
            pickle.dump(minmax_graph_feature, f)
            pickle.dump(dataset, f)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import shutil
import math
import torch
import numpy
from sklearn.neighbors import KNeighborsRegressor


def deterministic_graph_data(
    path: str,
    number_configurations: int = 500,
    configuration_start: int = 0,
    unit_cell_x_range: list = [1, 3],
    unit_cell_y_range: list = [1, 3],
    unit_cell_z_range: list = [1, 2],
    number_types: int = 3,
    types: list = None,
    number_neighbors: int = 2,
    linear_only=False,
):
    if types == None:
        types = range(number_types)

    # We assume that the unit cell is Body Center Cubic (BCC)
    unit_cell_x = torch.randint(
        unit_cell_x_range[0],
        unit_cell_x_range[1],
        (number_configurations,),
    )
    unit_cell_y = torch.randint(
        unit_cell_y_range[0],
        unit_cell_y_range[1],
        (number_configurations,),
    )
    unit_cell_z = torch.randint(
        unit_cell_z_range[0],
        unit_cell_z_range[1],
        (number_configurations,),
    )

    for configuration in range(number_configurations):
        uc_x = unit_cell_x[configuration]
        uc_y = unit_cell_y[configuration]
        uc_z = unit_cell_z[configuration]
        create_configuration(
            path,
            configuration,
            configuration_start,
            uc_x,
            uc_y,
            uc_z,
            types,
            number_neighbors,
            linear_only,
        )


def create_configuration(
    path,
    configuration,
    configuration_start,
    uc_x,
    uc_y,
    uc_z,
    types,
    number_neighbors,
    linear_only,
):
    ###############################################################################################
    ###################################   STRUCTURE OF THE DATA  ##################################
    ###############################################################################################

    #   GLOCAL_OUTPUT
    #   NODE1_FEATURE   NODE1_INDEX     NODE1_COORDINATE_X  NODE1_COORDINATE_Y  NODE1_COORDINATE_Z  NODAL_OUTPUT1   NODAL_OUTPUT2   NODAL_OUTPUT3
    #   NODE2_FEATURE   NODE2_INDEX     NODE2_COORDINATE_X  NODE2_COORDINATE_Y  NODE2_COORDINATE_Z  NODAL_OUTPUT1   NODAL_OUTPUT2   NODAL_OUTPUT3
    #   ...
    #   NODENn_FEATURE   NODEn_INDEX     NODEn_COORDINATE_X  NODEn_COORDINATE_Y  NODEn_COORDINATE_Z  NODAL_OUTPUT1   NODAL_OUTPUT2   NODAL_OUTPUT3

    ###############################################################################################
    #################################   FORMULAS FOR NODAL FEATURE  ###############################
    ###############################################################################################

    #   NODAL_FEATURE = MODULUS( NODE_INDEX, NUM_CLUSTERS )

    ###############################################################################################
    ##########################   FORMULAS FOR GLOBAL AND NODAL OUTOUTS  ###########################
    ###############################################################################################

    #   GLOBAL_OUTPUT = SUM_OVER_NODES ( NODAL_OUTPUT1 ) + SUM_OVER_NODES ( NODAL_OUTPUT2 ) + SUM_OVER_NODES ( NODAL_OUTPUT3 )
    #   NODAL_OUTPUT1(X) = X
    #   NODAL_OUTPUT2(X) = X^2
    #   NODAL_OUTPUT3(X) = X^3

    ###############################################################################################
    count_pos = 0
    number_nodes = 2 * uc_x * uc_y * uc_z
    positions = torch.zeros(number_nodes, 3)
    for x in range(uc_x):
        for y in range(uc_y):
            for z in range(uc_z):
                positions[count_pos][0] = x
                positions[count_pos][1] = y
                positions[count_pos][2] = z
                positions[count_pos + 1][0] = x + 0.5
                positions[count_pos + 1][1] = y + 0.5
                positions[count_pos + 1][2] = z + 0.5
                count_pos = count_pos + 2

    node_ids = torch.tensor(range(number_nodes), dtype=torch.int64).reshape(
        (number_nodes, 1)
    )
    node_feature = torch.randint(min(types), max(types) + 1, (number_nodes, 1))

    if linear_only:
        node_output_x = node_feature
    else:
        # We use a K nearest neighbor model to average nodal features and simulate a message passing between neighboring nodes
        knn = KNeighborsRegressor(number_neighbors)
        knn.fit(positions, node_feature)
        node_output_x = torch.Tensor(knn.predict(positions))

    node_output_x_square = node_output_x ** 2 + node_feature
    node_output_x_cube = node_output_x ** 3

    updated_table = torch.cat(
        (
            node_feature,
            node_ids,
            positions,
            node_output_x,
            node_output_x_square,
            node_output_x_cube,
        ),
        1,
    )
    updated_table = updated_table.detach().numpy()

    if linear_only:
        total_value = torch.sum(node_output_x)
    else:
        total_value_linear = torch.sum(node_output_x)
        total_value = (
            torch.sum(node_output_x)
            + torch.sum(node_output_x_square)
            + torch.sum(node_output_x_cube)
        )
    filetxt = numpy.array2string(total_value.detach().numpy())
    if not linear_only:
        filetxt += "\t" + numpy.array2string(total_value_linear.detach().numpy())

    for index in range(0, number_nodes):
        numpy_row = updated_table[index, :]
        numpy_string_row = numpy.array2string(
            numpy_row, precision=2, separator="\t", suppress_small=True
        )
        filetxt += "\n" + numpy_string_row.lstrip("[").rstrip("]")

    filename = os.path.join(
        path, "output" + str(configuration + configuration_start) + ".txt"
    )
    with open(filename, "w") as f:
        f.write(filetxt)



from abc import ABC, abstractmethod

import torch


class AbstractBaseDataset(torch.utils.data.Dataset, ABC):
    """
    HydraGNN's base dataset. This is abstract class.
    """

    def __init__(self):
        super().__init__()
        self.dataset = list()

    @abstractmethod
    def get(self, idx):
        """
        Return a dataset at idx
        """
        pass

    @abstractmethod
    def len(self):
        """
        Total number of dataset.
        If data is distributed, it should be the global total size.
        """
        pass

    def apply(self, func):
        for data in self.dataset:
            func(data)

    def map(self, func):
        for data in self.dataset:
            yield func(data)

    def __len__(self):
        return self.len()

    def __getitem__(self, idx):
        return self.get(idx)

    def __iter__(self):
        for idx in range(self.len()):
            yield self.get(idx)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import json, os
from functools import singledispatch

from hydragnn.preprocess.load_data import dataset_loading_and_splitting
from hydragnn.utils.distributed import setup_ddp, get_distributed_model
from hydragnn.utils.model import load_existing_model
from hydragnn.utils.config_utils import (
    update_config,
    get_log_name_config,
)
from hydragnn.models.create import create_model_config
from hydragnn.train.train_validate_test import test
from hydragnn.postprocess.postprocess import output_denormalize


@singledispatch
def run_prediction(config):
    raise TypeError("Input must be filename string or configuration dictionary.")


@run_prediction.register
def _(config_file: str):

    with open(config_file, "r") as f:
        config = json.load(f)

    run_prediction(config)


@run_prediction.register
def _(config: dict):

    try:
        os.environ["SERIALIZED_DATA_PATH"]
    except:
        os.environ["SERIALIZED_DATA_PATH"] = os.getcwd()

    world_size, world_rank = setup_ddp()

    train_loader, val_loader, test_loader = dataset_loading_and_splitting(config=config)

    config = update_config(config, train_loader, val_loader, test_loader)

    model = create_model_config(
        config=config["NeuralNetwork"], verbosity=config["Verbosity"]["level"]
    )

    model = get_distributed_model(
        model,
        config["Verbosity"]["level"],
        sync_batch_norm=config["NeuralNetwork"]["Architecture"]["SyncBatchNorm"],
    )

    log_name = get_log_name_config(config)
    load_existing_model(model, log_name)

    (
        error,
        error_rmse_task,
        true_values,
        predicted_values,
    ) = test(test_loader, model, config["Verbosity"]["level"])

    ##output predictions with unit/not normalized
    if config["NeuralNetwork"]["Variables_of_interest"]["denormalize_output"]:
        true_values, predicted_values = output_denormalize(
            config["NeuralNetwork"]["Variables_of_interest"]["y_minmax"],
            true_values,
            predicted_values,
        )

    return error, error_rmse_task, true_values, predicted_values



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import math
import collections
import torch
import sklearn

# function to return key for any value
def get_keys(dictionary, val):
    keys = []
    for key, value in dictionary.items():
        if val == value:
            keys.append(key)
    return keys


def get_max_graph_size(dataset):
    max_graph_size = int(0)
    for data in dataset:
        max_graph_size = max(max_graph_size, data.num_nodes)
    return max_graph_size


def get_elements_list(dataset):
    ## Identify all the elements present in at least one configuration
    elements_list = torch.FloatTensor()

    for data in dataset:
        elements = torch.unique(data.x[:, 0])
        elements_list = torch.cat((elements_list, elements), 0)

    elements_list = torch.unique(elements_list)

    return elements_list


def create_dictionary_from_elements_list(elements_list: list):
    dictionary = {}

    for index, element in enumerate(elements_list):
        dictionary[element.item()] = index

    return dictionary


def create_dataset_categories(dataset):
    max_graph_size = get_max_graph_size(dataset)
    power_ten = math.ceil(math.log10(max_graph_size))
    elements_list = get_elements_list(dataset)
    elements_dictionary = create_dictionary_from_elements_list(elements_list)

    dataset_categories = []

    for data in dataset:
        elements, frequencies = torch.unique(data.x[:, 0], return_counts=True)
        category = 0
        for element, frequency in zip(elements, frequencies):
            category += frequency.item() * (
                10 ** (power_ten * elements_dictionary[element.item()])
            )
        dataset_categories.append(category)

    return dataset_categories


def duplicate_unique_data_samples(dataset, dataset_categories):
    counter = collections.Counter(dataset_categories)
    keys = get_keys(counter, 1)
    augmented_data = []
    augmented_data_category = []

    for data, category in zip(dataset, dataset_categories):
        if category in keys:
            # Data augmentation on unique elements to allow additional splitting
            augmented_data.append(data.clone())
            augmented_data_category.append(category)

    if isinstance(dataset, list):
        dataset.extend(augmented_data)
    else:
        dataset.dataset.extend(augmented_data)
    dataset_categories.extend(augmented_data_category)

    return dataset, dataset_categories


def generate_partition(
    sss: sklearn.model_selection.StratifiedShuffleSplit, dataset, dataset_categories
):
    parition1_indices = []
    partition2_indices = []
    partition1_set = []
    partition2_set = []

    for train_index, val_test_index in sss.split(dataset, dataset_categories):
        parition1_indices = train_index.tolist()
        partition2_indices = val_test_index.tolist()

    for index in parition1_indices:
        partition1_set.append(dataset[index])

    for index in partition2_indices:
        partition2_set.append(dataset[index])

    return partition1_set, partition2_set


def compositional_stratified_splitting(dataset, perc_train):
    """Given the dataset and the percentage of data you want to extract from it, method will
    apply stratified sampling where X is the dataset and Y is are the category values for each datapoint.
    In the case each structure contains 2 types of atoms, the category will
    be constructed as such: number of atoms of type 1 + number of atoms of type 2 * 100.
    Parameters
    ----------
    dataset: [Data]
        A list of Data objects representing a structure that has atoms.
    subsample_percentage: float
        Percentage of the dataset.
    Returns
    ----------
    [Data]
        Subsample of the original dataset constructed using stratified sampling.
    """
    dataset_categories = create_dataset_categories(dataset)
    dataset, dataset_categories = duplicate_unique_data_samples(
        dataset, dataset_categories
    )

    sss_train = sklearn.model_selection.StratifiedShuffleSplit(
        n_splits=1, train_size=perc_train, random_state=0
    )
    trainset, val_test_set = generate_partition(sss_train, dataset, dataset_categories)

    val_test_dataset_categories = create_dataset_categories(val_test_set)
    val_test_set, val_test_dataset_categories = duplicate_unique_data_samples(
        val_test_set, val_test_dataset_categories
    )

    sss_valtest = sklearn.model_selection.StratifiedShuffleSplit(
        n_splits=1, train_size=0.5, random_state=0
    )
    valset, testset = generate_partition(
        sss_valtest, val_test_set, val_test_dataset_categories
    )

    return trainset, valset, testset



from .dataset_descriptors import AtomFeatures, StructureFeatures

from .utils import (
    check_if_graph_size_variable,
    check_if_graph_size_variable_dist,
    get_radius_graph,
    get_radius_graph_pbc,
    get_radius_graph_config,
    get_radius_graph_pbc_config,
    RadiusGraphPBC,
    update_predicted_values,
    update_atom_features,
    stratified_sampling,
)

from .load_data import (
    dataset_loading_and_splitting,
    create_dataloaders,
    split_dataset,
    transform_raw_data_to_serialized,
    total_to_train_val_test_pkls,
    HydraDataLoader,
)
from .serialized_dataset_loader import (
    SerializedDataLoader,
)
from .lsms_raw_dataset_loader import LSMS_RawDataLoader
from .cfg_raw_dataset_loader import CFG_RawDataLoader
from .compositional_data_splitting import compositional_stratified_splitting



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import time
import torch
from .distributed import get_comm_size_and_rank, get_device
from .print_utils import print_distributed


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class Timer:
    timers_local = dict()
    timers_min = dict()
    timers_max = dict()
    timers_avg = dict()
    number_calls = dict()

    def __init__(self, name: str):
        self.start_time = None
        self.elapsed_time = None
        self.tmin = None
        self.tmax = None
        self.tavg = None
        self.running = False
        self.calls = 0
        self.name = name

        self.world_size, self.world_rank = get_comm_size_and_rank()
        self.device = get_device()

        self.timers_local.setdefault(name, 0.0)
        self.timers_min.setdefault(name, 0.0)
        self.timers_max.setdefault(name, 0.0)
        self.timers_avg.setdefault(name, 0.0)
        self.number_calls.setdefault(name, 0)

    def start(self):
        if self.start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self.running = True
        self.calls = self.calls + 1
        self.start_time = time.perf_counter()

    def stop(self):
        if self.start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        self.elapsed_time = time.perf_counter() - self.start_time
        self.start_time = None

        self.tmin = torch.Tensor([self.elapsed_time]).to(self.device)
        self.tmax = torch.Tensor([self.elapsed_time]).to(self.device)
        self.tavg = torch.Tensor([self.elapsed_time]).to(self.device)

        world_size, world_rank = get_comm_size_and_rank()

        if torch.distributed.is_initialized():
            torch.distributed.reduce(self.tmin, 0, op=torch.distributed.ReduceOp.MIN)
            self.tmin = self.tmin.item()
            torch.distributed.reduce(self.tmax, 0, op=torch.distributed.ReduceOp.MAX)
            self.tmax = self.tmax.item()
            torch.distributed.reduce(self.tavg, 0, op=torch.distributed.ReduceOp.SUM)
        self.tavg = self.tavg.item() / world_size

        self.timers_local[self.name] += self.elapsed_time
        self.timers_min[self.name] += self.tmin
        self.timers_max[self.name] += self.tmax
        self.timers_avg[self.name] += self.tavg
        self.number_calls[self.name] += self.calls

        self.running = False

    def reset(self):
        self.start_time = None
        self.elapsed_time = None
        self.running = False
        self.tmin = None
        self.tmax = None
        self.tavg = None
        self.calls = 0


def print_timers(verbosity):

    world_size, world_rank = get_comm_size_and_rank()

    # With proper lever of verbosity >=1, the local timers will have different values per process
    [
        print_distributed(
            verbosity,
            f"Process {world_rank} - Local timer: ",
            key,
            " : ",
            round(value, 2),
        )
        for key, value in sorted(
            Timer.timers_local.items(), key=lambda item: item[1], reverse=True
        )
    ]

    # The statistics are the result of global collective operations, so we only print them once
    if verbosity >= 1:
        if world_size > 1:
            print_distributed(1, "Minimum timers: ")
            [
                print_distributed(1, key, " : ", round(value, 2))
                for key, value in sorted(
                    Timer.timers_min.items(), key=lambda item: item[1], reverse=True
                )
            ]
            print_distributed(1, "Maximum timers: ")
            [
                print_distributed(1, key, " : ", round(value, 2))
                for key, value in sorted(
                    Timer.timers_max.items(), key=lambda item: item[1], reverse=True
                )
            ]
            print_distributed(1, "Average timers: ")
            [
                print_distributed(1, key, " : ", round(value, 2))
                for key, value in sorted(
                    Timer.timers_avg.items(), key=lambda item: item[1], reverse=True
                )
            ]
        print_distributed(1, "Number of calls to timers: ")
        print_distributed(1, Timer.number_calls)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import socket

import torch
import torch.distributed as dist
import torch_geometric

# FIXME: deprecated in torch_geometric 2.0
try:
    from torch_geometric.loader import DataLoader
except:
    from torch_geometric.data import DataLoader

from hydragnn.preprocess.serialized_dataset_loader import SerializedDataLoader
from hydragnn.preprocess.lsms_raw_dataset_loader import LSMS_RawDataLoader
from hydragnn.preprocess.cfg_raw_dataset_loader import CFG_RawDataLoader
from hydragnn.preprocess.compositional_data_splitting import (
    compositional_stratified_splitting,
)
from hydragnn.utils.distributed import get_comm_size_and_rank
from hydragnn.utils.time_utils import Timer
import pickle

from hydragnn.utils.print_utils import print_master, log

from torch_geometric.data import Batch, Dataset
from torch.utils.data.dataloader import _DatasetKind

from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import multiprocessing as mp
import queue
import time
import sys
import re


def parse_omp_places(envstr):
    """
    Parse OMP_PLACES env string to get list of places
    Usage example:
        parse_omp_places(os.environ["OMP_PLACES"])
    Input examples:
        "{0:4},{4:4},{8:4},{12:4},{16:4},{20:4},{24:4}"
    """
    plist = list()
    for block in re.findall(r"({[\d,:]+})", envstr):
        start, cnt = list(map(int, re.findall(r"\d+", block)))
        for i in range(start, start + cnt):
            plist.append(i)
    return plist


class SimpleDataLoader(DataLoader):
    """
    A naive implementation of a custom dataloader
    """

    def __init__(self, dataset, **kwargs):
        super(HydraDataLoader, self).__init__(dataset, **kwargs)
        self._dataset_fetcher = _DatasetKind.create_fetcher(
            self._dataset_kind,
            self.dataset,
            self._auto_collation,
            self.collate_fn,
            self.drop_last,
        )

        log("num_workers:", self.num_workers)
        log("len:", len(self._index_sampler))

    def __iter__(self):
        self._num_yielded = 0
        self._sampler_iter = iter(self._index_sampler)
        return self

    def __next__(self):
        self._num_yielded += 1
        index = next(self._sampler_iter)
        data = self._dataset_fetcher.fetch(index)  # may raise StopIteration
        return data


class HydraDataLoader(DataLoader):
    """
    A custom data loader with multi-threading on a HPC system.
    This is to overcome a few problems (affinity, hanging, crashing, etc)
    with Pytorch's multi-threaded DataLoader on Summit and Perlmutter.
    (2022/08) jyc: This is a work-in-progress version. Performance is not verified.
    """

    def __init__(self, dataset, **kwargs):
        super(HydraDataLoader, self).__init__(dataset, **kwargs)
        self._dataset_fetcher = _DatasetKind.create_fetcher(
            self._dataset_kind,
            self.dataset,
            self._auto_collation,
            self.collate_fn,
            self.drop_last,
        )

        ## List of threads job (futures)
        self.fs = queue.Queue()

        log("num_workers:", self.num_workers)
        log("len:", len(self._index_sampler))

    @staticmethod
    def worker_init(counter):
        core_width = 2
        if os.getenv("HYDRAGNN_AFFINITY_WIDTH") is not None:
            core_width = int(os.environ["HYDRAGNN_AFFINITY_WIDTH"])

        core_offset = 0
        if os.getenv("HYDRAGNN_AFFINITY_OFFSET") is not None:
            core_offset = int(os.environ["HYDRAGNN_AFFINITY_OFFSET"])

        with counter.get_lock():
            wid = counter.value
            counter.value += 1

        affinity = None
        if hasattr(os, "sched_getaffinity"):
            affinity_check = os.getenv("HYDRAGNN_AFFINITY")
            if affinity_check == "OMP":
                affinity = parse_omp_places(os.getenv("OMP_PLACES"))
            else:
                affinity = list(os.sched_getaffinity(0))

            affinity_mask = set(
                affinity[
                    core_width * wid
                    + core_offset : core_width * (wid + 1)
                    + core_offset
                ]
            )
            os.sched_setaffinity(0, affinity_mask)
            affinity = os.sched_getaffinity(0)

        hostname = socket.gethostname()
        log(
            f"Worker: pid={os.getpid()} hostname={hostname} ID={wid} affinity={affinity}"
        )
        return 0

    @staticmethod
    def fetch(dataset, ibatch, index, pin_memory=False):
        batch = [dataset[i] for i in index]
        # hostname = socket.gethostname()
        # log (f"Worker done: pid={os.getpid()} hostname={hostname} ibatch={ibatch}")
        data = Batch.from_data_list(batch)
        if pin_memory:
            data = torch.utils.data._utils.pin_memory.pin_memory(data)
        return (ibatch, data)

    def __iter__(self):
        log("Iterator reset")
        ## Check previous futures
        if self.fs.qsize() > 0:
            log("Clearn previous futures:", self.fs.qsize())
            for future in iter(self.fs.get, None):
                future.cancel()

        ## Resetting
        self._num_yielded = 0
        self._sampler_iter = iter(self._index_sampler)
        self.fs_iter = iter(self.fs.get, None)
        counter = mp.Value("i", 0)
        executor = ThreadPoolExecutor(
            max_workers=self.num_workers,
            initializer=self.worker_init,
            initargs=(counter,),
        )
        for i in range(len(self._index_sampler)):
            index = next(self._sampler_iter)
            future = executor.submit(
                self.fetch,
                self.dataset,
                i,
                index,
                pin_memory=self.pin_memory,
            )
            self.fs.put(future)
        self.fs.put(None)
        # log ("Submit all done.")
        return self

    def __next__(self):
        # log ("Getting next", self._num_yielded)
        future = next(self.fs_iter)
        ibatch, data = future.result()
        # log (f"Future done: ibatch={ibatch}", data.num_graphs)
        self._num_yielded += 1
        return data


def dataset_loading_and_splitting(config: {}):
    ##check if serialized pickle files or folders for raw files provided
    if not list(config["Dataset"]["path"].values())[0].endswith(".pkl"):
        transform_raw_data_to_serialized(config["Dataset"])

    ##if total dataset is provided, split the dataset and save them to pkl files and update config with pkl file locations
    if "total" in config["Dataset"]["path"].keys():
        total_to_train_val_test_pkls(config)

    trainset, valset, testset = load_train_val_test_sets(config)

    return create_dataloaders(
        trainset,
        valset,
        testset,
        batch_size=config["NeuralNetwork"]["Training"]["batch_size"],
    )


def create_dataloaders(trainset, valset, testset, batch_size):
    if dist.is_initialized():

        train_sampler = torch.utils.data.distributed.DistributedSampler(trainset)
        val_sampler = torch.utils.data.distributed.DistributedSampler(valset)
        test_sampler = torch.utils.data.distributed.DistributedSampler(testset)

        pin_memory = True
        persistent_workers = False
        num_workers = 0
        if os.getenv("HYDRAGNN_NUM_WORKERS") is not None:
            num_workers = int(os.environ["HYDRAGNN_NUM_WORKERS"])

        use_custom_dataloader = 0
        if os.getenv("HYDRAGNN_CUSTOM_DATALOADER") is not None:
            use_custom_dataloader = int(os.environ["HYDRAGNN_CUSTOM_DATALOADER"])

        if use_custom_dataloader == 1:
            train_loader = HydraDataLoader(
                trainset,
                batch_size=batch_size,
                sampler=train_sampler,
                num_workers=num_workers,
                pin_memory=pin_memory,
                persistent_workers=persistent_workers,
            )
        else:
            train_loader = DataLoader(
                trainset,
                batch_size=batch_size,
                shuffle=False,
                sampler=train_sampler,
                num_workers=num_workers,
                pin_memory=pin_memory,
                persistent_workers=persistent_workers,
            )
        val_loader = DataLoader(
            valset, batch_size=batch_size, shuffle=False, sampler=val_sampler
        )
        test_loader = DataLoader(
            testset, batch_size=batch_size, shuffle=False, sampler=test_sampler
        )

    else:

        train_loader = DataLoader(trainset, batch_size=batch_size, shuffle=True)
        val_loader = DataLoader(
            valset,
            batch_size=batch_size,
            shuffle=True,
        )
        test_loader = DataLoader(
            testset,
            batch_size=batch_size,
            shuffle=True,
        )

    return train_loader, val_loader, test_loader


def split_dataset(
    dataset: [],
    perc_train: float,
    stratify_splitting: bool,
):
    if not stratify_splitting:
        perc_val = (1 - perc_train) / 2
        data_size = len(dataset)
        trainset = dataset[: int(data_size * perc_train)]
        valset = dataset[
            int(data_size * perc_train) : int(data_size * (perc_train + perc_val))
        ]
        testset = dataset[int(data_size * (perc_train + perc_val)) :]
    else:
        trainset, valset, testset = compositional_stratified_splitting(
            dataset, perc_train
        )

    return trainset, valset, testset


def load_train_val_test_sets(config, isdist=False):
    timer = Timer("load_data")
    timer.start()

    dataset_list = []
    datasetname_list = []

    for dataset_name, raw_data_path in config["Dataset"]["path"].items():
        if raw_data_path.endswith(".pkl"):
            files_dir = raw_data_path
        else:
            files_dir = f"{os.environ['SERIALIZED_DATA_PATH']}/serialized_dataset/{config['Dataset']['name']}_{dataset_name}.pkl"
        # loading serialized data and recalculating neighbourhoods depending on the radius and max num of neighbours
        loader = SerializedDataLoader(config, dist=isdist)
        dataset = loader.load_serialized_data(dataset_path=files_dir)

        dataset_list.append(dataset)
        datasetname_list.append(dataset_name)

    trainset = dataset_list[datasetname_list.index("train")]
    valset = dataset_list[datasetname_list.index("validate")]
    testset = dataset_list[datasetname_list.index("test")]

    timer.stop()

    return trainset, valset, testset


def transform_raw_data_to_serialized(config):
    _, rank = get_comm_size_and_rank()

    if rank == 0:
        if config["format"] == "LSMS" or config["format"] == "unit_test":
            loader = LSMS_RawDataLoader(config)
        elif config["format"] == "CFG":
            loader = CFG_RawDataLoader(config)
        else:
            raise NameError("Data format not recognized for raw data loader")

        loader.load_raw_data()

    if dist.is_initialized():
        dist.barrier()


def total_to_train_val_test_pkls(config, isdist=False):
    _, rank = get_comm_size_and_rank()

    if list(config["Dataset"]["path"].values())[0].endswith(".pkl"):
        file_dir = config["Dataset"]["path"]["total"]
    else:
        file_dir = f"{os.environ['SERIALIZED_DATA_PATH']}/serialized_dataset/{config['Dataset']['name']}.pkl"
    # if "total" raw dataset is provided, generate train/val/test pkl files and update config dict.
    with open(file_dir, "rb") as f:
        minmax_node_feature = pickle.load(f)
        minmax_graph_feature = pickle.load(f)
        dataset_total = pickle.load(f)

    trainset, valset, testset = split_dataset(
        dataset=dataset_total,
        perc_train=config["NeuralNetwork"]["Training"]["perc_train"],
        stratify_splitting=config["Dataset"]["compositional_stratified_splitting"],
    )
    serialized_dir = os.path.dirname(file_dir)
    config["Dataset"]["path"] = {}
    for dataset_type, dataset in zip(
        ["train", "validate", "test"], [trainset, valset, testset]
    ):
        serial_data_name = config["Dataset"]["name"] + "_" + dataset_type + ".pkl"
        config["Dataset"]["path"][dataset_type] = (
            serialized_dir + "/" + serial_data_name
        )
        if (not isdist) and (rank == 0):
            with open(os.path.join(serialized_dir, serial_data_name), "wb") as f:
                pickle.dump(minmax_node_feature, f)
                pickle.dump(minmax_graph_feature, f)
                pickle.dump(dataset, f)
        elif isdist:
            ## This is for the ising example.
            ## Each process writes own pickle data. config["Dataset"]["name"] contains rank info.
            with open(os.path.join(serialized_dir, serial_data_name), "wb") as f:
                pickle.dump(minmax_node_feature, f)
                pickle.dump(minmax_graph_feature, f)
                pickle.dump(dataset, f)

    if dist.is_initialized():
        dist.barrier()



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import matplotlib as mpl

mpl.use("Agg")
import matplotlib.pyplot as plt
from itertools import chain
import time, pickle
import numpy as np
from math import sqrt, floor, ceil

plt.rcParams.update({"font.size": 18})


class Visualizer:
    """A class used for visualizing results of GCNN outputs.

    Methods
    -------
    __init__(self, model_with_config_name: str, node_feature=[], num_heads=1, head_dims=[1])
        Create a Visualizer object with model name (also the location to save the plots), node feature list,
        number of heads, and head dimensions list.
    __hist2d_contour(self, data1, data2)
        Calculate joint histogram of data1 and data2 values.
    __err_condmean(self, data1, data2, weight=1.0)
        Calculate conditional mean values of the weighted difference between data1 and data2 on data1, i.e., <weight*|data1-data2|>_data1.
    __scatter_impl(self,ax,x,y,s=None,c=None,marker=None,title=None,x_label=None,y_label=None,xylim_equal=False,)
        Create scatter plot of x and y values.

    ##create plots for a single variable with name varname
    create_plot_global_analysis(self, varname, true_values, predicted_values, save_plot=True)
        Creates scatter/conditonal mean/error pdf plots from the true and predicted values of variable varname.
        For node level output, statistics across all nodes, l2 length and sum, are also considered for analysis.
        (file: varname +"_scatter_condm_err.png")
    create_parity_plot_and_error_histogram_scalar(self, varname, true_values, predicted_values, iepoch=None, save_plot=True)
        Creates scatter plots for true_values and predicted values and error histogram.
        (file: varname + ".png")
    create_parity_plot_vector(self, varname, true_values, predicted_values, iepoch=None, save_plot=True)
        Creates scatter plots for true_values and predicted values and error histogram.
        (file: varname + ".png")
    create_error_histogram_per_node(self, varname, true_values, predicted_values, iepoch=None, save_plot=True)
        Creates error histogram plot for the true and predicted values per node.
        (file: varname+"_error_hist1d_"+ str(iepoch).zfill(4)+ ".png" or varname+"_error_hist1d.png" )
    create_parity_plot_per_node_vector(self, varname, true_values, predicted_values, iepoch=None, save_plot=True):
        Creates scatter plots per node for true and predicted values of length 3 vector variable varname.
        (file: varname+"_"+ str(iepoch).zfill(4) + ".png" or varname+".png" )

    #create plots for all heads
    plot_history(self,total_loss_train,total_loss_val,total_loss_test,task_loss_train_sum,task_loss_val_sum, task_loss_test_sum,task_loss_train,task_loss_val,task_loss_test,task_weights,task_names,):
        Save history of losses to file and plot.
    create_scatter_plots(self, true_values, predicted_values, output_names=None, iepoch=None)
        Creates scatter plots for all head predictions. One plot for each head
    create_plot_global(self, true_values, predicted_values, output_names=None)
        Creates global analysis for all head predictons, e.g., scatter/condmean/error pdf plot. One plot for each head
    """

    def __init__(
        self,
        model_with_config_name: str,
        node_feature: [],
        num_nodes_list: [],
        num_heads=1,
        head_dims=[1],
    ):
        self.true_values = []
        self.predicted_values = []
        self.model_with_config_name = model_with_config_name
        self.node_feature = node_feature
        self.num_nodes = len(node_feature[0])
        self.num_heads = num_heads
        self.head_dims = head_dims
        self.num_nodes_list = num_nodes_list

    def __hist2d_contour(self, data1, data2):
        hist2d_pasr, xedge_pasr, yedge_pasr = np.histogram2d(
            np.hstack(data1), np.hstack(data2), bins=50
        )
        xcen_pasr = 0.5 * (xedge_pasr[0:-1] + xedge_pasr[1:])
        ycen_pasr = 0.5 * (yedge_pasr[0:-1] + yedge_pasr[1:])
        BCTY_pasr, BCTX_pasr = np.meshgrid(ycen_pasr, xcen_pasr)
        hist2d_pasr = hist2d_pasr / np.amax(hist2d_pasr)
        return BCTX_pasr, BCTY_pasr, hist2d_pasr

    def __err_condmean(self, data1, data2, weight=1.0):
        errabs = np.abs(np.hstack(data1) - np.hstack(data2)) * weight
        hist2d_pasr, xedge_pasr, yedge_pasr = np.histogram2d(
            np.hstack(data1), errabs, bins=50
        )
        xcen_pasr = 0.5 * (xedge_pasr[0:-1] + xedge_pasr[1:])
        ycen_pasr = 0.5 * (yedge_pasr[0:-1] + yedge_pasr[1:])
        hist2d_pasr = hist2d_pasr / np.amax(hist2d_pasr)
        mean1d_cond = np.dot(hist2d_pasr, ycen_pasr) / (
            np.sum(hist2d_pasr, axis=1) + 1e-12
        )
        return xcen_pasr, mean1d_cond

    def __scatter_impl(
        self,
        ax,
        x,
        y,
        s=None,
        c=None,
        marker=None,
        title=None,
        x_label=None,
        y_label=None,
        xylim_equal=False,
    ):

        ax.scatter(x, y, s=s, edgecolor="b", marker=marker, facecolor="none")

        ax.set_title(title + ", number of samples =" + str(len(x)))
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        if xylim_equal:
            ax.set_aspect("equal")
            minimum = np.min((ax.get_xlim(), ax.get_ylim()))
            maximum = np.max((ax.get_xlim(), ax.get_ylim()))
            ax.set_xlim(minimum, maximum)
            ax.set_ylim(minimum, maximum)

        self.add_identity(ax, color="r", ls="--")

    def create_plot_global_analysis(
        self, varname, true_values, predicted_values, save_plot=True
    ):
        """Creates scatter/condmean/error pdf plots for the true and predicted values of variable varname."""
        nshape = np.asarray(predicted_values).shape
        if nshape[1] == 1:
            fig, axs = plt.subplots(1, 3, figsize=(15, 4.5))
            plt.subplots_adjust(
                left=0.08, bottom=0.15, right=0.95, top=0.925, wspace=0.35, hspace=0.1
            )
            ax = axs[0]
            self.__scatter_impl(
                ax,
                true_values,
                predicted_values,
                title="Scalar output",
                x_label="True",
                y_label="Predicted",
                xylim_equal=True,
            )

            ax = axs[1]
            xtrue, error = self.__err_condmean(
                true_values, predicted_values, weight=1.0
            )
            ax.plot(xtrue, error, "ro")
            ax.set_title("Conditional mean abs. error")
            ax.set_xlabel("True")
            ax.set_ylabel("abs. error")

            ax = axs[2]
            hist1d, bin_edges = np.histogram(
                np.array(predicted_values) - np.array(true_values),
                bins=40,
                density=True,
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")
            ax.set_title("Scalar output: error PDF")
            ax.set_xlabel("Error")
            ax.set_ylabel("PDF")
        else:
            fig, axs = plt.subplots(3, 3, figsize=(18, 16))
            vlen_true = []
            vlen_pred = []
            vsum_true = []
            vsum_pred = []
            for isamp in range(nshape[0]):
                vlen_true.append(
                    sqrt(sum([comp ** 2 for comp in true_values[isamp][:]]))
                )
                vlen_pred.append(
                    sqrt(sum([comp ** 2 for comp in predicted_values[isamp][:]]))
                )
                vsum_true.append(sum(true_values[isamp][:]))
                vsum_pred.append(sum(predicted_values[isamp][:]))

            ax = axs[0, 0]
            self.__scatter_impl(
                ax,
                vlen_true,
                vlen_pred,
                title="Vector output: length",
                x_label="True",
                y_label="Predicted",
                xylim_equal=True,
            )

            ax = axs[1, 0]
            xtrue, error = self.__err_condmean(
                vlen_true, vlen_pred, weight=1.0 / sqrt(nshape[1])
            )
            ax.plot(xtrue, error, "ro")
            ax.set_ylabel("Conditional mean abs error")
            ax.set_xlabel("True")

            ax = axs[2, 0]
            hist1d, bin_edges = np.histogram(
                np.array(vlen_pred) - np.array(vlen_true), bins=40, density=True
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")
            ax.set_ylabel("Error PDF")
            ax.set_xlabel("Error")

            ax = axs[0, 1]
            self.__scatter_impl(
                ax,
                vsum_true,
                vsum_pred,
                title="Vector output: sum",
                x_label="True",
                y_label="Predicted",
                xylim_equal=True,
            )

            ax = axs[1, 1]
            xtrue, error = self.__err_condmean(
                vsum_true, vsum_pred, weight=1.0 / nshape[1]
            )
            ax.plot(xtrue, error, "ro")

            ax = axs[2, 1]
            hist1d, bin_edges = np.histogram(
                np.array(vsum_pred) - np.array(vsum_true), bins=40, density=True
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")

            truecomp = []
            predcomp = []
            for icomp in range(nshape[1]):
                truecomp.append(true_values[:][icomp])
                predcomp.append(predicted_values[:][icomp])

            ax = axs[0, 2]
            self.__scatter_impl(
                ax,
                truecomp,
                predcomp,
                title="Vector output: components",
                x_label="True",
                y_label="Predicted",
                xylim_equal=True,
            )

            ax = axs[1, 2]
            xtrue, error = self.__err_condmean(truecomp, predcomp, weight=1.0)
            ax.plot(xtrue, error, "ro")

            ax = axs[2, 2]
            hist1d, bin_edges = np.histogram(
                np.array(predcomp) - np.array(truecomp), bins=40, density=True
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")

            plt.subplots_adjust(
                left=0.075, bottom=0.1, right=0.98, top=0.95, wspace=0.2, hspace=0.25
            )

        if save_plot:
            fig.savefig(
                f"./logs/{self.model_with_config_name}/"
                + varname
                + "_scatter_condm_err.png"
            )
            plt.close()
        else:
            plt.show()

    def create_parity_plot_and_error_histogram_scalar(
        self, varname, true_values, predicted_values, iepoch=None, save_plot=True
    ):
        """Creates scatter plots for true_values and predicted values of variable varname at iepoch."""

        true_values = true_values.cpu().detach().numpy()
        predicted_values = predicted_values.cpu().detach().numpy()
        nshape = np.asarray(predicted_values).shape
        if nshape[1] == 1:
            fig, axs = plt.subplots(1, 2, figsize=(12, 6))
            ax = axs[0]
            self.__scatter_impl(
                ax,
                true_values,
                predicted_values,
                title=varname,
                x_label="True",
                y_label="Predicted",
                xylim_equal=True,
            )

            ax = axs[1]
            hist1d, bin_edges = np.histogram(
                predicted_values - true_values,
                bins=40,
                density=True,
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")
            ax.set_title(varname + ": error PDF")

            plt.subplots_adjust(
                left=0.075, bottom=0.1, right=0.98, top=0.9, wspace=0.2, hspace=0.25
            )
        else:
            nrow = floor(sqrt((nshape[1] + 2)))
            ncol = ceil((nshape[1] + 2) / nrow)
            fig, axs = plt.subplots(nrow, ncol, figsize=(ncol * 3, nrow * 3))
            axs = axs.flatten()
            for inode in range(nshape[1]):
                xfeature = []
                truecomp = []
                predcomp = []
                for isamp in range(nshape[0]):
                    xfeature.append(self.node_feature[isamp][inode])
                    truecomp.append(true_values[isamp][inode])
                    predcomp.append(predicted_values[isamp][inode])
                ax = axs[inode]
                self.__scatter_impl(
                    ax,
                    truecomp,
                    predcomp,
                    s=6,
                    c=xfeature,
                    title="node:" + str(inode),
                    xylim_equal=True,
                )

            ax = axs[nshape[1]]  # summation over all the nodes/nodes
            xfeature = []
            truecomp = []
            predcomp = []
            for isamp in range(nshape[0]):
                xfeature.append(sum(self.node_feature[isamp][:]))
                truecomp.append(sum(true_values[isamp][:]))
                predcomp.append(sum(predicted_values[isamp][:]))
            self.__scatter_impl(
                ax, truecomp, predcomp, s=40, c=xfeature, title="SUM", xylim_equal=True
            )
            # summation over all the samples for each node
            ax = axs[nshape[1] + 1]
            xfeature = []
            truecomp = []
            predcomp = []
            for inode in range(nshape[1]):
                xfeature.append(sum(self.node_feature[:][inode]))
                truecomp.append(sum(true_values[:][inode]))
                predcomp.append(sum(predicted_values[:][inode]))
            self.__scatter_impl(
                ax,
                truecomp,
                predcomp,
                s=40,
                c=xfeature,
                title="SMP_Mean4sites:0-" + str(nshape[1]),
                xylim_equal=True,
            )

            for iext in range(nshape[1] + 2, axs.size):
                axs[iext].axis("off")
            plt.subplots_adjust(
                left=0.05, bottom=0.05, right=0.98, top=0.95, wspace=0.1, hspace=0.25
            )
        if save_plot:
            if iepoch:
                fig.savefig(
                    f"./logs/{self.model_with_config_name}/"
                    + varname
                    + "_"
                    + str(iepoch).zfill(4)
                    + ".png"
                )
            else:
                fig.savefig(f"./logs/{self.model_with_config_name}/" + varname + ".png")
            plt.close()
        return

    def create_error_histogram_per_node(
        self, varname, true_values, predicted_values, iepoch=None, save_plot=True
    ):
        """Creates error histogram plot for true and predicted values of variable varname at iepoch.[node level]"""

        nshape = predicted_values.shape
        if nshape[1] == 1:
            return
        else:
            nrow = floor(sqrt((nshape[1] + 2)))
            ncol = ceil((nshape[1] + 2) / nrow)
            # error plots
            fig, axs = plt.subplots(nrow, ncol, figsize=(ncol * 3.5, nrow * 3.2))
            axs = axs.flatten()
            for inode in range(nshape[1]):
                xfeature = []
                truecomp = []
                predcomp = []
                for isamp in range(nshape[0]):
                    xfeature.append(self.node_feature[isamp][inode])
                    truecomp.append(true_values[isamp][inode])
                    predcomp.append(predicted_values[isamp][inode])
                hist1d, bin_edges = np.histogram(
                    np.array(predcomp) - np.array(truecomp), bins=40, density=True
                )
                ax = axs[inode]
                ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")
                ax.set_title("node:" + str(inode))

            ax = axs[nshape[1]]
            xfeature = []
            truecomp = []
            predcomp = []
            for isamp in range(nshape[0]):
                xfeature.append(sum(self.node_feature[isamp][:]))
                truecomp.append(sum(true_values[isamp][:]))
                predcomp.append(sum(predicted_values[isamp][:]))
            hist1d, bin_edges = np.histogram(
                np.array(predcomp) - np.array(truecomp), bins=40, density=True
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")
            ax.set_title("SUM")

            # summation over all the samples for each node
            ax = axs[nshape[1] + 1]
            xfeature = []
            truecomp = []
            predcomp = []
            for inode in range(nshape[1]):
                xfeature.append(sum(self.node_feature[:][inode]))
                truecomp.append(sum(true_values[:][inode]))
                predcomp.append(sum(predicted_values[:][inode]))
            hist1d, bin_edges = np.histogram(
                np.array(predcomp) - np.array(truecomp), bins=40, density=True
            )
            ax.plot(0.5 * (bin_edges[:-1] + bin_edges[1:]), hist1d, "ro")
            ax.set_title("SMP_Mean4sites:0-" + str(nshape[1]))

            for iext in range(nshape[1] + 2, axs.size):
                axs[iext].axis("off")
            plt.subplots_adjust(
                left=0.075, bottom=0.1, right=0.98, top=0.9, wspace=0.2, hspace=0.35
            )
            if save_plot:
                if iepoch:
                    fig.savefig(
                        f"./logs/{self.model_with_config_name}/"
                        + varname
                        + "_error_hist1d_"
                        + str(iepoch).zfill(4)
                        + ".png"
                    )
                else:
                    fig.savefig(
                        f"./logs/{self.model_with_config_name}/"
                        + varname
                        + "_error_hist1d.png"
                    )
                plt.close()

    def create_parity_plot_vector(
        self,
        varname,
        true_values,
        predicted_values,
        head_dim,
        iepoch=None,
        save_plot=True,
    ):
        """Creates scatter plots for true and predicted values of vector variable varname."""

        predicted_vec = np.reshape(np.asarray(predicted_values), (-1, head_dim))
        true_vec = np.reshape(np.asarray(true_values), (-1, head_dim))
        num_samples = true_vec.shape[0]

        markers_vec = ["o", "s", "d"]  # different markers for three vector components
        nrow = floor(sqrt(head_dim))
        ncol = ceil(head_dim / nrow)
        fig, axs = plt.subplots(nrow, ncol, figsize=(ncol * 4, nrow * 4))
        axs = axs.flatten()
        for icomp in range(head_dim):
            ax = axs[icomp]
            self.__scatter_impl(
                ax,
                true_vec[:, icomp],
                predicted_vec[:, icomp],
                s=6,
                c="b",
                marker=markers_vec[icomp],
                title="comp:" + str(icomp),
                xylim_equal=True,
            )
        for iext in range(head_dim, axs.size):
            axs[iext].axis("off")
        plt.subplots_adjust(
            left=0.1, bottom=0.12, right=0.98, top=0.9, wspace=0.2, hspace=0.25
        )

        if save_plot:
            if iepoch:
                fig.savefig(
                    f"./logs/{self.model_with_config_name}/"
                    + varname
                    + "_"
                    + str(iepoch).zfill(4)
                    + ".png"
                )
            else:
                fig.savefig(f"./logs/{self.model_with_config_name}/" + varname + ".png")
            plt.close()

    # FIXME: this function is currently unused and is explicitly written for 3d vectors.
    def create_parity_plot_per_node_vector(
        self, varname, true_values, predicted_values, iepoch=None, save_plot=True
    ):
        """Creates scatter plots for true and predicted values of vector variable varname[nodel level]."""

        nshape = np.asarray(predicted_values).shape
        predicted_vec = np.reshape(np.asarray(predicted_values), (nshape[0], -1, 3))
        true_vec = np.reshape(np.asarray(true_values), (nshape[0], -1, 3))
        num_nodes = true_vec.shape[1]

        markers_vec = ["o", "s", "d"]  # different markers for three vector components
        nrow = floor(sqrt((num_nodes + 2)))
        ncol = ceil((num_nodes + 2) / nrow)
        fig, axs = plt.subplots(nrow, ncol, figsize=(ncol * 3, nrow * 3))
        axs = axs.flatten()
        for inode in range(num_nodes):
            for icomp in range(3):
                xfeature = []
                truecomp = []
                predcomp = []
                for isamp in range(nshape[0]):
                    xfeature.append(self.node_feature[isamp][inode])
                    truecomp.append(true_vec[isamp, inode, icomp])
                    predcomp.append(predicted_vec[isamp, inode, icomp])
                ax = axs[inode]
                self.__scatter_impl(
                    ax,
                    truecomp,
                    predcomp,
                    s=6,
                    c=xfeature,
                    marker=markers_vec[icomp],
                    title="node:" + str(inode),
                    xylim_equal=True,
                )

        ax = axs[num_nodes]  # summation over all the nodes/nodes
        for icomp in range(3):
            xfeature = []
            truecomp = []
            predcomp = []
            for isamp in range(nshape[0]):
                xfeature.append(sum(self.node_feature[isamp][:]))
                truecomp.append(sum(true_vec[isamp, :, icomp]))
                predcomp.append(sum(predicted_vec[isamp, :, icomp]))
            self.__scatter_impl(
                ax,
                truecomp,
                predcomp,
                s=40,
                c=xfeature,
                marker=markers_vec[icomp],
                title="SUM",
                xylim_equal=True,
            )

        ax = axs[num_nodes + 1]  # summation over all the samples for each node
        for icomp in range(3):
            xfeature = []
            truecomp = []
            predcomp = []
            for inode in range(num_nodes):
                xfeature.append(sum(self.node_feature[:][inode]))
                truecomp.append(sum(true_vec[:, inode, icomp]))
                predcomp.append(sum(predicted_vec[:, inode, icomp]))
            self.__scatter_impl(
                ax,
                truecomp,
                predcomp,
                s=40,
                c=xfeature,
                marker=markers_vec[icomp],
                title="SMP_Mean4sites:0-" + str(num_nodes),
                xylim_equal=True,
            )

        for iext in range(num_nodes + 2, axs.size):
            axs[iext].axis("off")
        plt.subplots_adjust(
            left=0.05, bottom=0.05, right=0.98, top=0.95, wspace=0.1, hspace=0.25
        )

        if save_plot:
            if iepoch:
                fig.savefig(
                    f"./logs/{self.model_with_config_name}/"
                    + varname
                    + "_"
                    + str(iepoch).zfill(4)
                    + ".png"
                )
            else:
                fig.savefig(f"./logs/{self.model_with_config_name}/" + varname + ".png")
            plt.close()

    def add_identity(self, axes, *line_args, **line_kwargs):
        (identity,) = axes.plot([], [], *line_args, **line_kwargs)

        def callback(axes):
            low_x, high_x = axes.get_xlim()
            low_y, high_y = axes.get_ylim()
            low = max(low_x, low_y)
            high = min(high_x, high_y)
            identity.set_data([low, high], [low, high])

        callback(axes)
        axes.callbacks.connect("xlim_changed", callback)
        axes.callbacks.connect("ylim_changed", callback)
        return axes

    def plot_history(
        self,
        total_loss_train,
        total_loss_val,
        total_loss_test,
        task_loss_train,
        task_loss_val,
        task_loss_test,
        task_weights,
        task_names,
    ):
        nrow = 1
        fhist = open(f"./logs/{self.model_with_config_name}/history_loss.pckl", "wb")
        pickle.dump(
            [
                total_loss_train,
                total_loss_val,
                total_loss_test,
                task_loss_train,
                task_loss_val,
                task_loss_test,
                task_weights,
                task_names,
            ],
            fhist,
        )
        fhist.close()
        num_tasks = len(task_loss_train[0])
        if num_tasks > 0:
            task_loss_train = task_loss_train.cpu().detach().numpy()
            task_loss_val = task_loss_val.cpu().detach().numpy()
            task_loss_test = task_loss_test.cpu().detach().numpy()
            nrow = 2
        fig, axs = plt.subplots(nrow, num_tasks, figsize=(16, 6 * nrow))
        axs = axs.flatten()
        ax = axs[0]
        ax.plot(total_loss_train.cpu().detach().numpy(), "-", label="train")
        ax.plot(total_loss_val.cpu().detach().numpy(), ":", label="validation")
        ax.plot(total_loss_test.cpu().detach().numpy(), "--", label="test")
        ax.set_title("total loss")
        ax.set_xlabel("Epochs")
        ax.set_yscale("log")
        ax.legend()
        for iext in range(1, num_tasks):
            axs[iext].axis("off")
        for ivar in range(task_loss_train.shape[1]):
            ax = axs[num_tasks + ivar]
            ax.plot(task_loss_train[:, ivar], label="train")
            ax.plot(task_loss_val[:, ivar], label="validation")
            ax.plot(task_loss_test[:, ivar], "--", label="test")
            ax.set_title(task_names[ivar] + ", {:.4f}".format(task_weights[ivar]))
            ax.set_xlabel("Epochs")
            ax.set_yscale("log")
            if ivar == 0:
                ax.legend()
        for iext in range(num_tasks + task_loss_train.shape[1], axs.size):
            axs[iext].axis("off")
        plt.subplots_adjust(
            left=0.1, bottom=0.08, right=0.98, top=0.9, wspace=0.25, hspace=0.3
        )
        fig.savefig(f"./logs/{self.model_with_config_name}/history_loss.png")
        plt.close()

    def create_scatter_plots(
        self, true_values, predicted_values, output_names=None, iepoch=None
    ):
        """Creates scatter plots for all head predictions."""
        for ihead in range(self.num_heads):
            head_true = true_values[ihead].cpu()
            head_pred = predicted_values[ihead].cpu()
            if self.head_dims[ihead] > 1:
                # vector output
                self.create_parity_plot_vector(
                    output_names[ihead],
                    head_true,
                    head_pred,
                    self.head_dims[ihead],
                    iepoch,
                )
            else:
                self.create_parity_plot_and_error_histogram_scalar(
                    output_names[ihead],
                    head_true,
                    head_pred,
                    iepoch,
                )
                self.create_error_histogram_per_node(
                    output_names[ihead],
                    head_true,
                    head_pred,
                    iepoch,
                )

    def create_plot_global(self, true_values, predicted_values, output_names=None):
        """Creates global analysis for all head predictons, e.g., scatter/condmean/error pdf plot."""
        for ihead in range(self.num_heads):
            head_true = true_values[ihead].cpu()
            head_pred = predicted_values[ihead].cpu()
            self.create_plot_global_analysis(
                output_names[ihead],
                head_true,
                head_pred,
                save_plot=True,
            )

    def num_nodes_plot(
        self,
    ):
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        ax.hist(self.num_nodes_list)
        ax.set_title("Histogram of graph size in test set")
        ax.set_xlabel("number of nodes")
        fig.savefig(f"./logs/{self.model_with_config_name}/num_nodes.png")
        plt.close()



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

from tqdm import tqdm
import numpy as np

import torch

from hydragnn.preprocess.serialized_dataset_loader import SerializedDataLoader
from hydragnn.postprocess.postprocess import output_denormalize
from hydragnn.postprocess.visualizer import Visualizer
from hydragnn.utils.print_utils import print_distributed, iterate_tqdm, log
from hydragnn.utils.time_utils import Timer
from hydragnn.utils.profile import Profiler
from hydragnn.utils.distributed import get_device, print_peak_memory
from hydragnn.preprocess.load_data import HydraDataLoader
from hydragnn.utils.model import Checkpoint, EarlyStopping

import os

from torch.profiler import record_function
import contextlib
from unittest.mock import MagicMock
from hydragnn.utils.distributed import get_comm_size_and_rank
import torch.distributed as dist
import pickle

import hydragnn.utils.tracer as tr


def train_validate_test(
    model,
    optimizer,
    train_loader,
    val_loader,
    test_loader,
    writer,
    scheduler,
    config,
    model_with_config_name,
    verbosity=0,
    plot_init_solution=True,
    plot_hist_solution=False,
    create_plots=False,
):
    num_epoch = config["Training"]["num_epoch"]
    EarlyStop = (
        config["Training"]["EarlyStopping"]
        if "EarlyStopping" in config["Training"]
        else False
    )

    SaveCheckpoint = (
        config["Training"]["Checkpoint"]
        if "Checkpoint" in config["Training"]
        else False
    )

    device = get_device()
    # total loss tracking for train/vali/test
    total_loss_train = torch.zeros(num_epoch, device=device)
    total_loss_val = torch.zeros(num_epoch, device=device)
    total_loss_test = torch.zeros(num_epoch, device=device)
    # loss tracking for each head/task
    task_loss_train = torch.zeros((num_epoch, model.module.num_heads), device=device)
    task_loss_test = torch.zeros((num_epoch, model.module.num_heads), device=device)
    task_loss_val = torch.zeros((num_epoch, model.module.num_heads), device=device)

    # preparing for results visualization
    ## collecting node feature
    if create_plots:
        node_feature = []
        nodes_num_list = []
        ## (2022/05) : FIXME: using test_loader.datast caused a bottleneck for large data
        for data in iterate_tqdm(
            test_loader.dataset, verbosity, desc="Collecting node feature"
        ):
            node_feature.extend(data.x.tolist())
            nodes_num_list.append(data.num_nodes)

        visualizer = Visualizer(
            model_with_config_name,
            node_feature=node_feature,
            num_heads=model.module.num_heads,
            head_dims=model.module.head_dims,
            num_nodes_list=nodes_num_list,
        )
        visualizer.num_nodes_plot()

    if create_plots and plot_init_solution:  # visualizing of initial conditions
        _, _, true_values, predicted_values = test(test_loader, model, verbosity)
        visualizer.create_scatter_plots(
            true_values,
            predicted_values,
            output_names=config["Variables_of_interest"]["output_names"],
            iepoch=-1,
        )

    profiler = Profiler("./logs/" + model_with_config_name)
    if "Profile" in config:
        profiler.setup(config["Profile"])

    if EarlyStop:
        earlystopper = EarlyStopping()
        if "patience" in config["Training"]:
            earlystopper = EarlyStopping(patience=config["Training"]["patience"])

    if SaveCheckpoint:
        checkpoint = Checkpoint(name=model_with_config_name)
        if "checkpoint_warmup" in config["Training"]:
            checkpoint = Checkpoint(
                name=model_with_config_name,
                warmup=config["Training"]["checkpoint_warmup"],
            )

    timer = Timer("train_validate_test")
    timer.start()

    for epoch in range(0, num_epoch):
        profiler.set_current_epoch(epoch)
        for dataloader in [train_loader, val_loader, test_loader]:
            if getattr(dataloader.sampler, "set_epoch", None) is not None:
                dataloader.sampler.set_epoch(epoch)

        with profiler as prof:
            tr.enable()
            tr.start("train")
            train_loss, train_taskserr = train(
                train_loader, model, optimizer, verbosity, profiler=prof
            )
            tr.stop("train")
            tr.disable()
            if epoch == 0:
                tr.reset()

        val_loss, val_taskserr = validate(
            val_loader, model, verbosity, reduce_ranks=True
        )
        test_loss, test_taskserr, true_values, predicted_values = test(
            test_loader,
            model,
            verbosity,
            reduce_ranks=True,
            return_samples=plot_hist_solution,
        )
        scheduler.step(val_loss)
        if writer is not None:
            writer.add_scalar("train error", train_loss, epoch)
            writer.add_scalar("validate error", val_loss, epoch)
            writer.add_scalar("test error", test_loss, epoch)
            for ivar in range(model.module.num_heads):
                writer.add_scalar(
                    "train error of task" + str(ivar), train_taskserr[ivar], epoch
                )
        print_distributed(
            verbosity,
            f"Epoch: {epoch:02d}, Train Loss: {train_loss:.8f}, Val Loss: {val_loss:.8f}, "
            f"Test Loss: {test_loss:.8f}",
        )
        print_distributed(
            verbosity, "Tasks Loss:", [taskerr.item() for taskerr in train_taskserr]
        )

        total_loss_train[epoch] = train_loss
        total_loss_val[epoch] = val_loss
        total_loss_test[epoch] = test_loss
        task_loss_train[epoch, :] = train_taskserr
        task_loss_val[epoch, :] = val_taskserr
        task_loss_test[epoch, :] = test_taskserr

        ###tracking the solution evolving with training
        if plot_hist_solution:
            visualizer.create_scatter_plots(
                true_values,
                predicted_values,
                output_names=config["Variables_of_interest"]["output_names"],
                iepoch=epoch,
            )

        if SaveCheckpoint:
            if checkpoint(model, optimizer, reduce_values_ranks(val_loss).item()):
                print_distributed(
                    verbosity, "Creating Checkpoint: %f" % checkpoint.min_perf_metric
                )
            print_distributed(
                verbosity, "Best Performance Metric: %f" % checkpoint.min_perf_metric
            )

        if EarlyStop:
            if earlystopper(reduce_values_ranks(val_loss)):
                print_distributed(
                    verbosity,
                    "Early stopping executed at epoch = %d due to val_loss not decreasing"
                    % epoch,
                )
                break

    timer.stop()

    if create_plots:
        # reduce loss statistics across all processes
        total_loss_train = reduce_values_ranks(total_loss_train)
        total_loss_val = reduce_values_ranks(total_loss_val)
        total_loss_test = reduce_values_ranks(total_loss_test)
        task_loss_train = reduce_values_ranks(task_loss_train)
        task_loss_val = reduce_values_ranks(task_loss_val)
        task_loss_test = reduce_values_ranks(task_loss_test)

        # At the end of training phase, do the one test run for visualizer to get latest predictions
        test_loss, test_taskserr, true_values, predicted_values = test(
            test_loader, model, verbosity
        )

        ##output predictions with unit/not normalized
        if config["Variables_of_interest"]["denormalize_output"]:
            true_values, predicted_values = output_denormalize(
                config["Variables_of_interest"]["y_minmax"],
                true_values,
                predicted_values,
            )

    _, rank = get_comm_size_and_rank()
    if create_plots and rank == 0:
        ######result visualization######
        visualizer.create_plot_global(
            true_values,
            predicted_values,
            output_names=config["Variables_of_interest"]["output_names"],
        )
        visualizer.create_scatter_plots(
            true_values,
            predicted_values,
            output_names=config["Variables_of_interest"]["output_names"],
        )
        ######plot loss history#####
        visualizer.plot_history(
            total_loss_train,
            total_loss_val,
            total_loss_test,
            task_loss_train,
            task_loss_val,
            task_loss_test,
            model.module.loss_weights,
            config["Variables_of_interest"]["output_names"],
        )


def get_head_indices(model, data):
    """In data.y (the true value here), all feature variables for a mini-batch are concatenated together as a large list.
    To calculate loss function, we need to know true value for each feature in every head.
    This function is to get the feature/head index/location in the large list."""
    if all(ele == "graph" for ele in model.module.head_type):
        return get_head_indices_graph(model, data)
    else:
        return get_head_indices_node_or_mixed(model, data)


def get_head_indices_graph(model, data):
    """this is for cases when outputs are all at graph level"""
    # total length
    nsize = data.y.shape[0]
    # feature index for all heads
    head_index = [None] * model.module.num_heads
    if model.module.num_heads == 1:
        head_index[0] = torch.arange(nsize)
        return head_index
    # dimensions of all heads
    head_dims = model.module.head_dims
    head_dimsum = sum(head_dims)

    batch_size = data.batch.max() + 1
    for ihead in range(model.module.num_heads):
        head_each = torch.arange(head_dims[ihead])
        head_ind_temporary = head_each.repeat(batch_size)
        head_shift_temporary = sum(head_dims[:ihead]) + torch.repeat_interleave(
            torch.arange(batch_size) * head_dimsum, head_dims[ihead]
        )
        head_index[ihead] = head_ind_temporary + head_shift_temporary
    return head_index


def get_head_indices_node_or_mixed(model, data):
    """this is for cases when outputs are node level or mixed graph-node level"""
    batch_size = data.batch.max() + 1
    y_loc = data.y_loc
    # head size for each sample
    total_size = y_loc[:, -1]
    # feature index for all heads
    head_index = [None] * model.module.num_heads
    if model.module.num_heads == 1:
        head_index[0] = torch.arange(data.y.shape[0])
        return head_index
    # intermediate work list
    head_ind_temporary = [None] * batch_size
    # track the start loc of each sample
    sample_start = torch.cumsum(total_size, dim=0) - total_size
    sample_start = sample_start.view(-1, 1)
    # shape (batch_size, model.module.num_heads), start and end of each head for each sample
    start_index = sample_start + y_loc[:, :-1]
    end_index = sample_start + y_loc[:, 1:]

    # a large index tensor pool for all element in data.y
    index_range = torch.arange(0, end_index[-1, -1], device=y_loc.device)
    for ihead in range(model.module.num_heads):
        for isample in range(batch_size):
            head_ind_temporary[isample] = index_range[
                start_index[isample, ihead] : end_index[isample, ihead]
            ]
        head_index[ihead] = torch.cat(head_ind_temporary, dim=0)

    return head_index


@torch.no_grad()
def reduce_values_ranks_dist(local_tensor):
    if dist.get_world_size() > 1:
        dist.all_reduce(local_tensor, op=dist.ReduceOp.SUM)
        local_tensor = local_tensor / dist.get_world_size()
    return local_tensor


@torch.no_grad()
def reduce_values_ranks_mpi(local_tensor):
    if dist.get_world_size() > 1:
        from mpi4py import MPI

        local_tensor = MPI.COMM_WORLD.allreduce(
            local_tensor.detach().cpu().numpy(), op=MPI.SUM
        )
        local_tensor = torch.tensor(local_tensor) / dist.get_world_size()
    return local_tensor


def reduce_values_ranks(local_tensor):
    backend = os.getenv("HYDRAGNN_AGGR_BACKEND", "torch")
    if backend == "mpi":
        return reduce_values_ranks_mpi(local_tensor)
    else:
        return reduce_values_ranks_dist(local_tensor)


@torch.no_grad()
def gather_tensor_ranks(head_values):
    if dist.get_world_size() > 1:
        head_values = head_values.to(get_device())
        size_local = torch.tensor(
            [head_values.shape[0]], dtype=torch.int64, device=head_values.device
        )
        size_all = [torch.ones_like(size_local) for _ in range(dist.get_world_size())]
        dist.all_gather(size_all, size_local)
        size_all = torch.cat(size_all, 0)
        max_size = size_all.max()

        padded = torch.empty(
            max_size,
            *head_values.shape[1:],
            dtype=head_values.dtype,
            device=head_values.device,
        )
        padded[: head_values.shape[0]] = head_values

        tensor_list = [torch.ones_like(padded) for _ in range(dist.get_world_size())]
        dist.all_gather(tensor_list, padded)
        tensor_list = torch.cat(tensor_list, 0)

        head_values = torch.zeros(
            size_all.sum(),
            *head_values.shape[1:],
            dtype=head_values.dtype,
            device=head_values.device,
        )
        for i, size in enumerate(size_all):
            start_idx = i * max_size
            end_idx = start_idx + size.item()
            if end_idx > start_idx:
                head_values[
                    size_all[:i].sum() : size_all[:i].sum() + size.item()
                ] = tensor_list[start_idx:end_idx]

    return head_values


def train(
    loader,
    model,
    opt,
    verbosity,
    profiler=None,
):
    if profiler is None:
        profiler = Profiler()

    total_error = torch.tensor(0.0, device=get_device())
    tasks_error = torch.zeros(model.module.num_heads, device=get_device())
    num_samples_local = 0
    model.train()

    use_ddstore = (
        hasattr(loader.dataset, "ddstore")
        and hasattr(loader.dataset.ddstore, "epoch_begin")
        and bool(int(os.getenv("HYDRAGNN_USE_ddstore", "0")))
    )
    tr.start("dataload")
    if use_ddstore:
        loader.dataset.ddstore.epoch_begin()
    for data in iterate_tqdm(loader, verbosity, desc="Train"):
        if use_ddstore:
            loader.dataset.ddstore.epoch_end()
        tr.stop("dataload")
        tr.start("zero_grad")
        with record_function("zero_grad"):
            opt.zero_grad()
        tr.stop("zero_grad")
        tr.start("get_head_indices")
        with record_function("get_head_indices"):
            head_index = get_head_indices(model, data)
        tr.stop("get_head_indices")
        tr.start("forward")
        with record_function("forward"):
            data = data.to(get_device())
            pred = model(data)
            loss, tasks_loss = model.module.loss(pred, data.y, head_index)
        tr.stop("forward")
        tr.start("backward")
        with record_function("backward"):
            loss.backward()
        tr.stop("backward")
        tr.start("opt_step")
        # print_peak_memory(verbosity, "Max memory allocated before optimizer step")
        opt.step()
        # print_peak_memory(verbosity, "Max memory allocated after optimizer step")
        tr.stop("opt_step")
        profiler.step()
        with torch.no_grad():
            total_error += loss * data.num_graphs
            num_samples_local += data.num_graphs
            for itask in range(len(tasks_loss)):
                tasks_error[itask] += tasks_loss[itask] * data.num_graphs
        tr.start("dataload")
        if use_ddstore:
            loader.dataset.ddstore.epoch_begin()
    if use_ddstore:
        loader.dataset.ddstore.epoch_end()
    tr.stop("dataload")

    train_error = total_error / num_samples_local
    tasks_error = tasks_error / num_samples_local
    return train_error, tasks_error


@torch.no_grad()
def validate(loader, model, verbosity, reduce_ranks=True):

    total_error = torch.tensor(0.0, device=get_device())
    tasks_error = torch.zeros(model.module.num_heads, device=get_device())
    num_samples_local = 0
    model.eval()
    use_ddstore = (
        hasattr(loader.dataset, "ddstore")
        and hasattr(loader.dataset.ddstore, "epoch_begin")
        and bool(int(os.getenv("HYDRAGNN_USE_ddstore", "0")))
    )
    if use_ddstore:
        loader.dataset.ddstore.epoch_begin()
    for data in iterate_tqdm(loader, verbosity, desc="Validate"):
        if use_ddstore:
            loader.dataset.ddstore.epoch_end()
        head_index = get_head_indices(model, data)
        data = data.to(get_device())
        pred = model(data)
        error, tasks_loss = model.module.loss(pred, data.y, head_index)
        total_error += error * data.num_graphs
        num_samples_local += data.num_graphs
        for itask in range(len(tasks_loss)):
            tasks_error[itask] += tasks_loss[itask] * data.num_graphs
        if use_ddstore:
            loader.dataset.ddstore.epoch_begin()
    if use_ddstore:
        loader.dataset.ddstore.epoch_end()

    val_error = total_error / num_samples_local
    tasks_error = tasks_error / num_samples_local
    if reduce_ranks:
        val_error = reduce_values_ranks(val_error)
        tasks_error = reduce_values_ranks(tasks_error)
    return val_error, tasks_error


@torch.no_grad()
def test(loader, model, verbosity, reduce_ranks=True, return_samples=True):

    total_error = torch.tensor(0.0, device=get_device())
    tasks_error = torch.zeros(model.module.num_heads, device=get_device())
    num_samples_local = 0
    model.eval()
    use_ddstore = (
        hasattr(loader.dataset, "ddstore")
        and hasattr(loader.dataset.ddstore, "epoch_begin")
        and bool(int(os.getenv("HYDRAGNN_USE_ddstore", "0")))
    )
    if use_ddstore:
        loader.dataset.ddstore.epoch_begin()
    for data in iterate_tqdm(loader, verbosity, desc="Test"):
        if use_ddstore:
            loader.dataset.ddstore.epoch_end()
        head_index = get_head_indices(model, data)
        data = data.to(get_device())
        pred = model(data)
        error, tasks_loss = model.module.loss(pred, data.y, head_index)
        total_error += error * data.num_graphs
        num_samples_local += data.num_graphs
        for itask in range(len(tasks_loss)):
            tasks_error[itask] += tasks_loss[itask] * data.num_graphs
        if use_ddstore:
            loader.dataset.ddstore.epoch_begin()
    if use_ddstore:
        loader.dataset.ddstore.epoch_end()

    test_error = total_error / num_samples_local
    tasks_error = tasks_error / num_samples_local
    true_values = [[] for _ in range(model.module.num_heads)]
    predicted_values = [[] for _ in range(model.module.num_heads)]
    if return_samples:
        for data in loader:
            head_index = get_head_indices(model, data)
            data = data.to(get_device())
            ytrue = data.y
            pred = model(data)
            for ihead in range(model.module.num_heads):
                head_pre = pred[ihead].reshape(-1, 1)
                head_val = ytrue[head_index[ihead]]
                true_values[ihead].append(head_val)
                predicted_values[ihead].append(head_pre)
        for ihead in range(model.module.num_heads):
            predicted_values[ihead] = torch.cat(predicted_values[ihead], dim=0)
            true_values[ihead] = torch.cat(true_values[ihead], dim=0)

    if reduce_ranks:
        test_error = reduce_values_ranks(test_error)
        tasks_error = reduce_values_ranks(tasks_error)
        if len(true_values[0]) > 0:
            for ihead in range(model.module.num_heads):
                true_values[ihead] = gather_tensor_ranks(true_values[ihead])
                predicted_values[ihead] = gather_tensor_ranks(predicted_values[ihead])

    return test_error, tasks_error, true_values, predicted_values



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import numpy as np
import pickle

import torch
from torch_geometric.data import Data
from torch import tensor

from hydragnn.utils.distributed import get_device
from hydragnn.utils.print_utils import log
from hydragnn.utils import nsplit, tensor_divide, comm_reduce

import random


class AbstractRawDataLoader:
    """A class used for loading raw files that contain data representing atom structures, transforms it and stores the structures as file of serialized structures.
    Most of the class methods are hidden, because from outside a caller needs only to know about
    load_raw_data method.

    Methods
    -------
    load_raw_data()
        Loads the raw files from specified path, performs the transformation to Data objects and normalization of values.
    """

    def __init__(self, config, dist=False):
        """
        config:
          shows the dataset path the target variables information, e.g, location and dimension, in data file
        ###########
        dataset_list:
          list of datasets read from self.path_dictionary
        serial_data_name_list:
          list of pkl file names
        node_feature_dim:
          list of dimensions of node features
        node_feature_col:
          list of column location/index (start location if dim>1) of node features
        graph_feature_dim:
          list of dimensions of graph features
        graph_feature_col: list,
          list of column location/index (start location if dim>1) of graph features

        dist: True if RawDataLoder is distributed (i.e., each RawDataLoader will read different subset of data)
        """

        self.dataset_list = []
        self.serial_data_name_list = []
        self.node_feature_name = (
            config["node_features"]["name"]
            if config["node_features"]["name"] is not None
            else None
        )
        self.node_feature_dim = config["node_features"]["dim"]
        self.node_feature_col = config["node_features"]["column_index"]
        self.graph_feature_name = (
            config["graph_features"]["name"]
            if config["graph_features"]["name"] is not None
            else None
        )
        self.graph_feature_dim = config["graph_features"]["dim"]
        self.graph_feature_col = config["graph_features"]["column_index"]
        self.raw_dataset_name = config["name"]
        self.data_format = config["format"]
        self.path_dictionary = config["path"]

        assert len(self.node_feature_name) == len(self.node_feature_dim)
        assert len(self.node_feature_name) == len(self.node_feature_col)
        assert len(self.graph_feature_name) == len(self.graph_feature_dim)
        assert len(self.graph_feature_name) == len(self.graph_feature_col)

        self.dist = dist
        if self.dist:
            assert torch.distributed.is_initialized()
            self.world_size = torch.distributed.get_world_size()
            self.rank = torch.distributed.get_rank()

    def load_raw_data(self):
        """Loads the raw files from specified path, performs the transformation to Data objects and normalization of values.
        After that the serialized data is stored to the serialized_dataset directory.
        """

        serialized_dir = os.environ["SERIALIZED_DATA_PATH"] + "/serialized_dataset"
        if not os.path.exists(serialized_dir):
            os.makedirs(serialized_dir, exist_ok=True)

        for dataset_type, raw_data_path in self.path_dictionary.items():
            if not os.path.isabs(raw_data_path):
                raw_data_path = os.path.join(os.getcwd(), raw_data_path)
            if not os.path.exists(raw_data_path):
                raise ValueError("Folder not found: ", raw_data_path)

            dataset = []
            assert (
                len(os.listdir(raw_data_path)) > 0
            ), "No data files provided in {}!".format(raw_data_path)

            filelist = sorted(os.listdir(raw_data_path))
            if self.dist:
                ## Random shuffle filelist to avoid the same test/validation set
                random.seed(43)
                random.shuffle(filelist)

                x = torch.tensor(len(filelist), requires_grad=False).to(get_device())
                y = x.clone().detach().requires_grad_(False)
                torch.distributed.all_reduce(x, op=torch.distributed.ReduceOp.MAX)
                assert x == y
                filelist = list(nsplit(filelist, self.world_size))[self.rank]
                log("local filelist", len(filelist))

            for name in filelist:
                if name == ".DS_Store":
                    continue
                # if the directory contains file, iterate over them
                if os.path.isfile(os.path.join(raw_data_path, name)):
                    data_object = self.transform_input_to_data_object_base(
                        filepath=os.path.join(raw_data_path, name)
                    )
                    if not isinstance(data_object, type(None)):
                        dataset.append(data_object)
                # if the directory contains subdirectories, explore their content
                elif os.path.isdir(os.path.join(raw_data_path, name)):
                    dir_name = os.path.join(raw_data_path, name)
                    for subname in os.listdir(dir_name):
                        if os.path.isfile(os.path.join(dir_name, subname)):
                            data_object = self.transform_input_to_data_object_base(
                                filepath=os.path.join(dir_name, subname)
                            )
                            if not isinstance(data_object, type(None)):
                                dataset.append(data_object)

            # scaled features by number of nodes
            dataset = self.scale_features_by_num_nodes(dataset)

            if dataset_type == "total":
                serial_data_name = self.raw_dataset_name + ".pkl"
            else:
                # append for train; test; validation
                serial_data_name = self.raw_dataset_name + "_" + dataset_type + ".pkl"

            self.dataset_list.append(dataset)
            self.serial_data_name_list.append(serial_data_name)

        self.normalize_dataset()

        for serial_data_name, dataset_normalized in zip(
            self.serial_data_name_list, self.dataset_list
        ):
            with open(os.path.join(serialized_dir, serial_data_name), "wb") as f:
                pickle.dump(self.minmax_node_feature, f)
                pickle.dump(self.minmax_graph_feature, f)
                pickle.dump(dataset_normalized, f)

    def transform_input_to_data_object_base(self, filepath):
        pass

    def scale_features_by_num_nodes(self, dataset):
        """Calculate [**]_scaled_num_nodes"""
        scaled_graph_feature_index = [
            i
            for i in range(len(self.graph_feature_name))
            if "_scaled_num_nodes" in self.graph_feature_name[i]
        ]
        scaled_node_feature_index = [
            i
            for i in range(len(self.node_feature_name))
            if "_scaled_num_nodes" in self.node_feature_name[i]
        ]

        for idx, data_object in enumerate(dataset):
            if dataset[idx].y is not None:
                dataset[idx].y[scaled_graph_feature_index] = (
                    dataset[idx].y[scaled_graph_feature_index] / data_object.num_nodes
                )
            if dataset[idx].x is not None:
                dataset[idx].x[:, scaled_node_feature_index] = (
                    dataset[idx].x[:, scaled_node_feature_index] / data_object.num_nodes
                )

        return dataset

    def normalize_dataset(self):

        """Performs the normalization on Data objects and returns the normalized dataset."""
        num_node_features = len(self.node_feature_dim)
        num_graph_features = len(self.graph_feature_dim)

        self.minmax_graph_feature = np.full((2, num_graph_features), np.inf)
        # [0,...]:minimum values; [1,...]: maximum values
        self.minmax_node_feature = np.full((2, num_node_features), np.inf)
        self.minmax_graph_feature[1, :] *= -1
        self.minmax_node_feature[1, :] *= -1
        for dataset in self.dataset_list:
            for data in dataset:
                # find maximum and minimum values for graph level features
                g_index_start = 0
                for ifeat in range(num_graph_features):
                    g_index_end = g_index_start + self.graph_feature_dim[ifeat]
                    self.minmax_graph_feature[0, ifeat] = min(
                        torch.min(data.y[g_index_start:g_index_end]),
                        self.minmax_graph_feature[0, ifeat],
                    )
                    self.minmax_graph_feature[1, ifeat] = max(
                        torch.max(data.y[g_index_start:g_index_end]),
                        self.minmax_graph_feature[1, ifeat],
                    )
                    g_index_start = g_index_end

                # find maximum and minimum values for node level features
                n_index_start = 0
                for ifeat in range(num_node_features):
                    n_index_end = n_index_start + self.node_feature_dim[ifeat]
                    self.minmax_node_feature[0, ifeat] = min(
                        torch.min(data.x[:, n_index_start:n_index_end]),
                        self.minmax_node_feature[0, ifeat],
                    )
                    self.minmax_node_feature[1, ifeat] = max(
                        torch.max(data.x[:, n_index_start:n_index_end]),
                        self.minmax_node_feature[1, ifeat],
                    )
                    n_index_start = n_index_end

        ## Gather minmax in parallel
        if self.dist:
            self.minmax_graph_feature[0, :] = comm_reduce(
                self.minmax_graph_feature[0, :], torch.distributed.ReduceOp.MIN
            )
            self.minmax_graph_feature[1, :] = comm_reduce(
                self.minmax_graph_feature[1, :], torch.distributed.ReduceOp.MAX
            )
            self.minmax_node_feature[0, :] = comm_reduce(
                self.minmax_node_feature[0, :], torch.distributed.ReduceOp.MIN
            )
            self.minmax_node_feature[1, :] = comm_reduce(
                self.minmax_node_feature[1, :], torch.distributed.ReduceOp.MAX
            )

        for dataset in self.dataset_list:
            for data in dataset:
                g_index_start = 0
                for ifeat in range(num_graph_features):
                    g_index_end = g_index_start + self.graph_feature_dim[ifeat]
                    data.y[g_index_start:g_index_end] = tensor_divide(
                        (
                            data.y[g_index_start:g_index_end]
                            - self.minmax_graph_feature[0, ifeat]
                        ),
                        (
                            self.minmax_graph_feature[1, ifeat]
                            - self.minmax_graph_feature[0, ifeat]
                        ),
                    )
                    g_index_start = g_index_end
                n_index_start = 0
                for ifeat in range(num_node_features):
                    n_index_end = n_index_start + self.node_feature_dim[ifeat]
                    data.x[:, n_index_start:n_index_end] = tensor_divide(
                        (
                            data.x[:, n_index_start:n_index_end]
                            - self.minmax_node_feature[0, ifeat]
                        ),
                        (
                            self.minmax_node_feature[1, ifeat]
                            - self.minmax_node_feature[0, ifeat]
                        ),
                    )
                    n_index_start = n_index_end



from mpi4py import MPI
import numpy as np

import torch
import torch_geometric.data

from hydragnn.utils.abstractbasedataset import AbstractBaseDataset

try:
    import pyddstore as dds
except ImportError:
    pass

from hydragnn.utils.print_utils import log
from hydragnn.utils import nsplit

import hydragnn.utils.tracer as tr


class DistDataset(AbstractBaseDataset):
    """Distributed dataset class"""

    def __init__(self, data, label, comm=MPI.COMM_WORLD, ddstore_width=None):
        super().__init__()

        self.label = label
        self.comm = comm
        self.rank = self.comm.Get_rank()
        self.comm_size = self.comm.Get_size()
        self.ddstore_width = (
            ddstore_width if ddstore_width is not None else self.comm_size
        )
        self.ddstore_comm = self.comm.Split(self.rank // self.ddstore_width, self.rank)
        self.ddstore_comm_rank = self.ddstore_comm.Get_rank()
        self.ddstore_comm_size = self.ddstore_comm.Get_size()
        self.ddstore = dds.PyDDStore(self.ddstore_comm)

        ## set total before set subset
        self.total_ns = len(data)
        rx = list(nsplit(range(len(data)), self.ddstore_comm_size))[
            self.ddstore_comm_rank
        ]
        for i in rx:
            self.dataset.append(data[i])

        self.keys = sorted(self.dataset[0].keys)
        self.variable_shape = dict()
        self.variable_dim = dict()
        self.variable_dtype = dict()
        self.variable_count = dict()
        self.variable_offset = dict()
        self.data = dict()
        nbytes = 0
        for k in self.keys:
            arr_list = [data[k].cpu().numpy() for data in self.dataset]
            m0 = np.min([x.shape for x in arr_list], axis=0)
            m1 = np.max([x.shape for x in arr_list], axis=0)
            vdims = list()
            for i in range(len(m0)):
                if m0[i] != m1[i]:
                    vdims.append(i)
            ## We can handle only single variable dimension.
            assert len(vdims) < 2
            vdim = 0
            if len(vdims) > 0:
                vdim = vdims[0]
            val = np.concatenate(arr_list, axis=vdim)
            assert val.data.contiguous

            self.variable_shape[k] = val.shape
            self.variable_dim[k] = vdim
            self.variable_dtype[k] = val.dtype

            vcount = np.array([x.shape[vdim] for x in arr_list])
            assert len(vcount) == len(self.dataset)
            vcount_list = self.ddstore_comm.allgather(vcount)
            vcount = np.hstack(vcount_list)
            self.variable_count[k] = vcount

            offset_arr = np.zeros_like(vcount)
            offset_arr[1:] = np.cumsum(vcount)[:-1]
            self.variable_offset[k] = offset_arr

            if vdim > 0:
                val = np.moveaxis(val, vdim, 0)
                val = np.ascontiguousarray(val)
            assert val.data.contiguous
            self.data[k] = val

            vname = "%s/%s" % (label, k)
            self.ddstore.add(vname, val)
            log(
                "DDStore add:",
                (
                    vname,
                    vdim,
                    val.dtype,
                    val.shape,
                    val.sum(),
                    (val.size * val.itemsize) / 1024 / 1024 / 1024,
                ),
            )
            nbytes += val.size * val.itemsize
        log("DDStore total (GB):", nbytes / 1024 / 1024 / 1024)

    def len(self):
        return self.total_ns

    @tr.profile("get")
    def get(self, idx):
        data_object = torch_geometric.data.Data()
        for k in self.keys:
            count = list(self.variable_shape[k])
            vdim = self.variable_dim[k]
            dtype = self.variable_dtype[k]
            offset = self.variable_offset[k][idx]
            count[vdim] = self.variable_count[k][idx]
            val = np.zeros(count, dtype=dtype)
            ## vdim should be the first dim for DDStore
            if vdim > 0:
                val = np.moveaxis(val, vdim, 0)
                val = np.ascontiguousarray(val)
                assert val.data.contiguous
            vname = "%s/%s" % (self.label, k)
            self.ddstore.get(vname, val, offset)
            if vdim > 0:
                val = np.moveaxis(val, 0, vdim)
                val = np.ascontiguousarray(val)
            v = torch.tensor(val)
            exec("data_object.%s = v" % (k))
        return data_object



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import sys, os, json
import pytest

import torch
from torch_geometric.data import Data
from torch_geometric.transforms import Distance, NormalizeRotation
from hydragnn.preprocess.utils import get_radius_graph_config

from hydragnn.preprocess.utils import (
    check_data_samples_equivalence,
)


def create_bct_sample():
    # Create BCT lattice structure with 32 nodes
    uc_x = 4
    uc_y = 2
    uc_z = 2
    lxy = 5.218
    lz = 7.058
    count = 0
    number_nodes = 2 * uc_x * uc_y * uc_z
    positions = torch.zeros(number_nodes, 3)
    for x in range(uc_x):
        for y in range(uc_y):
            for z in range(uc_z):
                positions[count][0] = x * lxy
                positions[count][1] = y * lxy
                positions[count][2] = z * lz
                count += 1
                positions[count][0] = (x + 0.5) * lxy
                positions[count][1] = (y + 0.5) * lxy
                positions[count][2] = (z + 0.5) * lz
                count += 1

    data = Data()
    data.pos = positions
    return data


def check_rotational_invariance(
    data, compute_edges, compute_edge_lengths, compute_rotation, tolerance
):
    # Create a copy of the same data sample that is going to be rotated
    data_rotated = data.clone()

    data = compute_edges(data)
    data = compute_edge_lengths(data)

    # Rotate clone data
    data_rotated = compute_rotation(data_rotated)

    data_rotated = compute_edges(data_rotated)
    data_rotated = compute_edge_lengths(data_rotated)

    assert check_data_samples_equivalence(data, data_rotated, tolerance)


def unittest_rotational_invariance(tol=1e-14):
    config_file = "./tests/inputs/ci_rotational_invariance.json"
    with open(config_file, "r") as f:
        config = json.load(f)
    compute_edges = get_radius_graph_config(config["Architecture"], loop=False)
    compute_edge_lengths = Distance(norm=False, cat=True)
    compute_rotation = NormalizeRotation(max_points=-1, sort=False)

    # Run first with single BCT structure.
    data = create_bct_sample()

    # Create random nodal features
    data.x = torch.randn(32, 1)

    # Create arbitrary global feature
    data.y = torch.tensor([[99]])

    check_rotational_invariance(
        data, compute_edges, compute_edge_lengths, compute_rotation, tol
    )

    # Repeat the same check on 10 graphs randomly generated
    for num_random_graphs in range(10):
        data = Data()

        # The position of the nodes in the graph is random
        data.pos = 3 * torch.randn(10, 3)

        # Create random nodal features
        data.x = torch.randn_like(data.pos)

        # Create arbitrary global feature
        data.y = torch.randn(1, 1)

        check_rotational_invariance(
            data, compute_edges, compute_edge_lengths, compute_rotation, tol
        )


@pytest.mark.mpi_skip()
def pytest_rotational_invariance():
    # Test with (default) single precision
    unittest_rotational_invariance(tol=1e-4)

    # Test with double precision
    torch.set_default_tensor_type(torch.DoubleTensor)
    unittest_rotational_invariance()



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

from torch_geometric.data import Data
from torch import tensor

from hydragnn.preprocess.raw_dataset_loader import AbstractRawDataLoader

# WARNING: DO NOT use collective communication calls here because only rank 0 uses this routines


class LSMS_RawDataLoader(AbstractRawDataLoader):
    """A class used for loading raw files that contain data representing atom structures, transforms it and stores the structures as file of serialized structures.
    Most of the class methods are hidden, because from outside a caller needs only to know about
    load_raw_data method.

    Methods
    -------
    load_raw_data()
        Loads the raw files from specified path, performs the transformation to Data objects and normalization of values.
    """

    def __init__(self, config, dist=False):
        super(LSMS_RawDataLoader, self).__init__(config, dist)

    def transform_input_to_data_object_base(self, filepath):
        data_object = self.__transform_LSMS_input_to_data_object_base(filepath=filepath)

        return data_object

    def __transform_LSMS_input_to_data_object_base(self, filepath):
        """Transforms lines of strings read from the raw data LSMS file to Data object and returns it.

        Parameters
        ----------
        lines:
          content of data file with all the graph information
        Returns
        ----------
        Data
            Data object representing structure of a graph sample.
        """

        data_object = Data()

        f = open(filepath, "r", encoding="utf-8")

        lines = f.readlines()
        graph_feat = lines[0].split(None, 2)
        g_feature = []
        # collect graph features
        for item in range(len(self.graph_feature_dim)):
            for icomp in range(self.graph_feature_dim[item]):
                it_comp = self.graph_feature_col[item] + icomp
                g_feature.append(float(graph_feat[it_comp].strip()))
        data_object.y = tensor(g_feature)

        node_feature_matrix = []
        node_position_matrix = []
        for line in lines[1:]:
            node_feat = line.split(None, 11)

            x_pos = float(node_feat[2].strip())
            y_pos = float(node_feat[3].strip())
            z_pos = float(node_feat[4].strip())
            node_position_matrix.append([x_pos, y_pos, z_pos])

            node_feature = []
            for item in range(len(self.node_feature_dim)):
                for icomp in range(self.node_feature_dim[item]):
                    it_comp = self.node_feature_col[item] + icomp
                    node_feature.append(float(node_feat[it_comp].strip()))
            node_feature_matrix.append(node_feature)

        f.close()

        data_object.pos = tensor(node_position_matrix)
        data_object.x = tensor(node_feature_matrix)
        data_object = self.__charge_density_update_for_LSMS(data_object)
        return data_object

    def __charge_density_update_for_LSMS(self, data_object: Data):
        """Calculate charge density for LSMS format
        Parameters
        ----------
        data_object: Data
            Data object representing structure of a graph sample.

        Returns
        ----------
        Data
            Data object representing structure of a graph sample.
        """
        num_of_protons = data_object.x[:, 0]
        charge_density = data_object.x[:, 1]
        charge_density -= num_of_protons
        data_object.x[:, 1] = charge_density
        return data_object



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import torch
import torch.nn.functional as F
from torch.nn import ModuleList
from torch.nn import ReLU, Linear
from torch_geometric.nn import GATv2Conv, BatchNorm, Sequential

from .Base import Base


class GATStack(Base):
    def __init__(
        self,
        heads: int,
        negative_slope: float,
        *args,
        **kwargs,
    ):
        # note that self.heads is a parameter in GATConv, not the num_heads in the output part
        self.heads = heads
        self.negative_slope = negative_slope

        super().__init__(*args, **kwargs)

    def _init_conv(self):
        """Here this function overwrites _init_conv() in Base since it has different implementation
        in terms of dimensions due to the multi-head attention"""
        self.graph_convs.append(self.get_conv(self.input_dim, self.hidden_dim, True))
        self.feature_layers.append(BatchNorm(self.hidden_dim * self.heads))
        for _ in range(self.num_conv_layers - 2):
            conv = self.get_conv(self.hidden_dim * self.heads, self.hidden_dim, True)
            self.graph_convs.append(conv)
            self.feature_layers.append(BatchNorm(self.hidden_dim * self.heads))
        conv = self.get_conv(self.hidden_dim * self.heads, self.hidden_dim, False)
        self.graph_convs.append(conv)
        self.feature_layers.append(BatchNorm(self.hidden_dim))

    def _init_node_conv(self):
        """Here this function overwrites _init_conv() in Base since it has different implementation
        in terms of dimensions due to the multi-head attention"""
        # *******convolutional layers for node level predictions*******#
        # two ways to implement node features from here:
        # 1. one graph for all node features
        # 2. one graph for one node features (currently implemented)
        if (
            "node" not in self.config_heads
            or self.config_heads["node"]["type"] != "conv"
        ):
            return
        node_feature_ind = [
            i for i, head_type in enumerate(self.head_type) if head_type == "node"
        ]
        if len(node_feature_ind) == 0:
            return
        # In this part, each head has same number of convolutional layers, but can have different output dimension
        self.convs_node_hidden.append(
            self.get_conv(self.hidden_dim, self.hidden_dim_node[0], True)
        )
        self.batch_norms_node_hidden.append(
            BatchNorm(self.hidden_dim_node[0] * self.heads)
        )
        for ilayer in range(self.num_conv_layers_node - 1):
            self.convs_node_hidden.append(
                self.get_conv(
                    self.hidden_dim_node[ilayer] * self.heads,
                    self.hidden_dim_node[ilayer + 1],
                    True,
                )
            )
            self.batch_norms_node_hidden.append(
                BatchNorm(self.hidden_dim_node[ilayer + 1] * self.heads)
            )
        for ihead in node_feature_ind:
            self.convs_node_output.append(
                self.get_conv(
                    self.hidden_dim_node[-1] * self.heads, self.head_dims[ihead], False
                )
            )
            self.batch_norms_node_output.append(BatchNorm(self.head_dims[ihead]))

    def get_conv(self, input_dim, output_dim, concat):
        gat = GATv2Conv(
            in_channels=input_dim,
            out_channels=output_dim,
            heads=self.heads,
            negative_slope=self.negative_slope,
            dropout=self.dropout,
            add_self_loops=True,
            concat=concat,
        )

        input_args = "x, pos, edge_index"
        conv_args = "x, edge_index"

        if self.use_edge_attr:
            input_args += ", edge_attr"
            conv_args += ", edge_attr"

        return Sequential(
            input_args,
            [
                (gat, conv_args + " -> x"),
                (lambda x, pos: [x, pos], "x, pos -> x, pos"),
            ],
        )

    def __str__(self):
        return "GATStack"
